import pytest
from scanpy.datasets import pbmc68k_reduced
import numpy as np

from liana.method._pipe_utils._pre import assert_covered, prep_check_adata, format_vars

adata = pbmc68k_reduced()
adata.layers['scaled_counts'] = adata.X


def test_prep_check_adata():
    temp = prep_check_adata(adata=adata, groupby='bulk_labels', min_cells=0,
                            use_raw=True, layer=None)
    np.testing.assert_almost_equal(np.sum(temp.X.data), 319044.22, decimal=2)

    desired = np.array([2.177, 2.177, 2.544, 2.544, 1.591, 1.591, 1.591, 1.591, 1.591, 1.591])
    np.testing.assert_almost_equal(temp.X.data[0:10], desired, decimal=3)

    # test filtering
    filt = prep_check_adata(adata=adata, groupby='bulk_labels',
                            min_cells=20, use_raw=True)
    assert len(filt.obs['label']) == 660


def test_check_if_covered():
    with pytest.raises(ValueError):
        assert_covered(['NOT', 'HERE'], adata.var_names, verbose=True)


def test_format_vars():
    a = ['CD4B_', 'CD8A', 'IL6']
    assert (np.array_equal(['CD4B', 'CD8A', 'IL6'], format_vars(a)))


def test_choose_mtx():
    # check if default is used correctly
    raw_adata = prep_check_adata(adata=adata, groupby='bulk_labels', min_cells=5)
    assert np.min(raw_adata.X.data) < 0

    # check if correct layer is returned
    by_layer = prep_check_adata(adata=adata, groupby='bulk_labels',
                                min_cells=5, use_raw=True)
    by_layer.layers['scaled'] = by_layer.X
    extracted = prep_check_adata(by_layer, groupby='bulk_labels',
                                 min_cells=5, layer='scaled')

    np.testing.assert_almost_equal(by_layer.layers['scaled'].data,
                                   extracted.X.data)


def test_choose_mtx_failure():
    # check exception if both layer and use_raw are provided
    with pytest.raises(ValueError):
        prep_check_adata(adata=adata, groupby='bulk_labels', min_cells=5,
                         layer='scaled_counts', use_raw=True)

    # check exception if .raw is not initialized
    adata.raw = None
    with pytest.raises(ValueError):
        prep_check_adata(adata=adata, groupby='bulk_labels', min_cells=5, use_raw=True)



from liana.method._Method import Method, MethodMeta
from .._pipe_utils._get_mean_perms import _get_lr_pvals


def _simple_mean(x, y): return (x + y) / 2


# Internal Function to calculate CellPhoneDB LR_mean and p-values
def _cpdb_score(x, perms, ligand_pos, receptor_pos, labels_pos) -> tuple:
    """
    Calculate CellPhoneDB-like LR means and p-values
    
    Parameters
    ----------
    x
        DataFrame row
    perms
        3D tensor with permuted averages per cluster
    ligand_pos
        Index of the ligand in the tensor
    receptor_pos
        Index of the receptor in the perms tensor
    labels_pos
        Index of cell identities in the perms tensor

    Returns
    -------
    A tuple with lr_mean and pvalue for x

    """
    if (x.ligand_means == 0) | (x.receptor_means == 0):
        return 0, 1

    return _get_lr_pvals(x, perms, ligand_pos, receptor_pos, labels_pos, _simple_mean)


# Initialize CPDB Meta
_cellphonedb = MethodMeta(method_name="CellPhoneDB",
                          complex_cols=["ligand_means", "receptor_means"],
                          add_cols=[],
                          fun=_cpdb_score,
                          magnitude="lr_means",
                          magnitude_ascending=False,
                          specificity="cellphone_pvals",
                          specificity_ascending=True,
                          permute=True,
                          reference="Efremova, M., Vento-Tormo, M., Teichmann, S.A. and "
                                    "Vento-Tormo, R., 2020. CellPhoneDB: inferring cell–cell "
                                    "communication from combined expression of multi-subunit "
                                    "ligand–receptor complexes. Nature protocols, 15(4), "
                                    "pp.1484-1506. "
                          )

# Initialize callable Method instance
cellphonedb = Method(_SCORE=_cellphonedb)



from liana.method._Method import Method, MethodMeta
from ._cellphonedb import _simple_mean


def _connectome_score(x) -> tuple:
    """
    Calculate Connectome-like Score

    Parameters
    ----------
    x
        DataFrame row

    Returns
    -------
    tuple(expr_prod, scaled_weight)

    """
    # magnitude
    expr_prod = x.ligand_means * x.receptor_means
    # specificity
    scaled_weight = _simple_mean(x.ligand_zscores, x.receptor_zscores)
    return expr_prod, scaled_weight


# Initialize CPDB Meta
_connectome = MethodMeta(method_name="Connectome",
                         complex_cols=['ligand_means', 'receptor_means'],
                         add_cols=['ligand_zscores', 'receptor_zscores'],
                         fun=_connectome_score,
                         magnitude='expr_prod',
                         magnitude_ascending=False,
                         specificity='scaled_weight',
                         specificity_ascending=False,
                         permute=False,
                         reference='Raredon, M.S.B., Yang, J., Garritano, J., Wang, M., Kushnir, '
                                   'D., Schupp, J.C., Adams, T.S., Greaney, A.M., Leiby, K.L., '
                                   'Kaminski, N. and Kluger, Y., 2022. Computation and '
                                   'visualization of cell–cell signaling topologies in '
                                   'single-cell systems data using Connectome. Scientific '
                                   'reports, 12(1), pp.1-12. '
                         )

# Initialize callable Method instance
connectome = Method(_SCORE=_connectome)



from liana.method._Method import Method, MethodMeta


def _natmi_score(x) -> tuple:
    """
    Calculate NATMI-like expression product and specificity weights

    Parameters
    ----------
    x
        DataFrame row

    Returns
    -------
    tuple(expr_prod, spec_weight)

    """

    # magnitude
    expr_prod = x.ligand_means * x.receptor_means

    # specificity
    s_lig = (x.ligand_means / x.ligand_means_sums)
    s_rec = (x.receptor_means / x.receptor_means_sums)
    spec_weight = s_lig * s_rec

    return expr_prod, spec_weight


# Initialize CPDB Meta
_natmi = MethodMeta(method_name="NATMI",
                    complex_cols=['ligand_means', 'receptor_means'],
                    add_cols=['ligand_means_sums', 'receptor_means_sums'],
                    fun=_natmi_score,
                    magnitude='expr_prod',
                    magnitude_ascending=False,
                    specificity='spec_weight',
                    specificity_ascending=False,
                    permute=False,
                    reference='Hou, R., Denisenko, E., Ong, H.T., Ramilowski, J.A. and Forrest, '
                              'A.R., 2020. Predicting cell-to-cell communication networks using '
                              'NATMI. Nature communications, 11(1), pp.1-11. '
                    )

# Initialize callable Method instance
natmi = Method(_SCORE=_natmi)




from types import ModuleType

import numpy as np
import pandas as pd
import scanpy as sc

from anndata import AnnData
from tqdm import tqdm

from ._common import _process_scores


def _check_if_mudata() -> ModuleType:

    try:
        from mudata import MuData

    except Exception:

        raise ImportError(
            'mudata is not installed. Please install it with: '
            'pip install mudata'
        )

    return MuData

## TODO generalize these functions to work with any package
def _check_if_decoupler() -> ModuleType:

    try:
        import decoupler as dc

    except Exception:

        raise ImportError(
            'decoupler is not installed. Please install it with: '
            'pip install decoupler'
        )

    return dc


def adata_to_views(adata, groupby, sample_key, obs_keys, view_separator=':', verbose=False, **kwargs):
    """
    Converts an AnnData object to a MuData object with views that represent an aggregate for each entity in `adata.obs[groupby]`.
    
    Parameters
    ----------
    adata: :class:`~anndata.AnnData`
        AnnData object
    groupby:
        Column name in `adata.obs` to group by
    sample_key:
        Column name in `adata.obs` to use as sample key
    obs_keys:
        Column names in `adata.obs` to merge with the MuData object
    view_separator
        Separator to use when assigning `adata.var_names` to views
    verbose
        If True, show progress bar.
    **kwargs
        Keyword arguments used to aggregate the values per cell into views. See `dc.get_pseudobulk` for more details.
    
    Returns
    -------
    Returns a MuData object with views that represent an aggregate for each entity in `adata.obs[groupby]`.
    
    """
    
    # Check if MuData & decoupler are installed
    MuData = _check_if_mudata()
    dc = _check_if_decoupler()
    
    views = adata.obs[groupby].unique()
    views = tqdm(views, disable=not verbose)
    
    padatas = {}
    for view in (views):
        # filter AnnData to view
        temp = adata[adata.obs[groupby] == view]
        # assign view to var_names
        temp.var_names = view + view_separator + temp.var_names
        
        padata = dc.get_pseudobulk(temp,
                                   sample_col=sample_key,
                                   groups_col=None, 
                                   **kwargs
                                   )
        
        # only append views that pass QC
        if 0 not in padata.shape:
            del padata.obs
            padatas[view] = padata
            
    # Convert to MuData
    mdata = MuData(padatas)
    
    # process metadata
    _process_meta(adata=adata, mdata=mdata, sample_key=sample_key, obs_keys=obs_keys)
    
    return mdata


def lrs_to_views(adata, 
                 score_key=None, 
                 inverse_fun= lambda x: 1 - x,
                 obs_keys=None,
                 lr_prop=0.5,
                 lr_fill=np.nan,
                 lrs_per_view=20,
                 lrs_per_sample=10,
                 samples_per_view=3,
                 min_variance=0,
                 lr_separator='^',
                 cell_separator='&',
                 var_separator=':',
                 uns_key = 'liana_res',
                 sample_key='sample',
                 source_key='source',
                 target_key='target', 
                 ligand_key='ligand_complex',
                 receptor_key='receptor_complex',
                 verbose=False
                 ):
    """
    Converts a LIANA result to a MuData object with views that represent an aggregate for each entity in `adata.obs[groupby]`.
    
    Parameters
    ----------
    adata
        AnnData object with LIANA results in `adata.uns[uns_key]`
    score_key
        Column in `adata.uns[uns_key]` that contains the scores to be used for building the views.
    inverse_fun
        Function that is applied to the scores before building the views. Default is `lambda x: 1 - x` which is used to invert the scores
        reflect probabilities (e.g. magnitude_rank), i.e. such for which lower values reflect higher relevance.
        This is handled automatically for the scores in liana.
    obs_keys
        List of keys in `adata.obs` that should be included in the MuData object. Default is `None`. 
        These columns should correspond to the number of samples in `adata.obs[sample_key]`.
    lr_prop
        Reflects the minimum required proportion of samples for an interaction to be considered for building the views. Default is `0.5`.
    lr_fill
        Value to fill in for interactions that are not present in a view. Default is `np.nan`.
    lrs_per_sample
        Reflects the minimum required number of interactions in a sample to be considered when building a specific view. Default is `10`.
    lrs_per_view
        Reflects the minimum required number of interactions in a view to be considered for building the views. Default is `20`.
    samples_per_view
        Reflects the minimum required samples to keep a view. Default is `3`.
    min_variance
        Reflects the minimum required variance across samples for each interaction in each view. Default is `0`.
        NaNs are ignored when computing the variance.
    lr_separator
        Separator to use for the interaction names in the views. Default is `^`.
    cell_separator
        Separator to use for the cell names in the views. Default is `&`.
    var_separator
        Separator to use for the variable names in the views. Default is `:`.
    uns_key
        Key in `adata.uns` that contains the LIANA results. Default is `'liana_res'`.
    sample_key
        Key in `adata.uns[uns_key]` that contains the sample names. Default is `'sample'`.
    source_key
        Key in `adata.uns[uns_key]` that contains the source names. Default is `'source'`.
    target_key
        Key in `adata.uns[uns_key]` that contains the target names. Default is `'target'`.
    ligand_key
        Key in `adata.uns[uns_key]` that contains the ligand names. Default is `'ligand_complex'`.
    receptor_key
        Key in `adata.uns[uns_key]` that contains the receptor names. Default is `'receptor_complex'`.
    verbose
        If True, show progress bar.
    
    Returns
    -------
    Returns a MuData object with views that represent an aggregate for each entity in `adata.obs[groupby]`.
    
    """
    
    # Check if MuData is installed
    MuData = _check_if_mudata()
    
    if (sample_key not in adata.obs.columns) or (sample_key not in adata.uns[uns_key].columns):
        raise ValueError(f'`{sample_key}` not found in `adata.obs` or `adata.uns[uns_key]`!' +
                         'Please ensure that the sample key is present in both objects.')
    
    if uns_key not in adata.uns_keys():
        raise ValueError(f'`{uns_key}` not found in `adata.uns`! Please run `li.mt.rank_aggregate.by_sample` first.')
    
    liana_res = adata.uns[uns_key].copy()
    
    if (score_key is None) or (score_key not in liana_res.columns):
        raise ValueError(f"Score column `{score_key}` not found in `liana_res`")
    
    if isinstance(obs_keys, list):
        if any([key not in adata.obs.keys() for key in obs_keys]):
            raise ValueError(f'`{obs_keys}` not found in `adata.obs`!')
    elif obs_keys is not None:
        raise ValueError(f'`obs_keys` must be a list or `None`!')
    
    keys = np.array([sample_key, source_key, target_key, ligand_key, receptor_key])
    missing_keys = keys[[ key not in liana_res.columns for key in keys]]
    
    if any(missing_keys):
        raise ValueError(f'`{missing_keys}` not found in `adata.uns[{uns_key}]`! Please check your input.')
    
    # concat columns (needed for MOFA)
    liana_res['interaction'] = liana_res[ligand_key] + lr_separator + liana_res[receptor_key]
    liana_res['ct_pair'] = liana_res[source_key] + cell_separator + liana_res[target_key]
    liana_res = liana_res[[sample_key, 'ct_pair', 'interaction', score_key]]
    
    # get scores & invert if necessary
    liana_res = _process_scores(liana_res=liana_res,
                                score_key=score_key,
                                inverse_fun=inverse_fun)
        
    # count samples per interaction
    count_pairs = (liana_res.
                   drop(columns=score_key).
                   groupby(['interaction', 'ct_pair']).
                   count().
                   rename(columns={sample_key: 'count'}).
                   reset_index()
                   )
    
    sample_n = liana_res[sample_key].nunique()
    
    # Keep only lrs above a certain proportion of samples
    count_pairs = count_pairs[count_pairs['count'] >= sample_n * lr_prop]
    liana_res = liana_res.merge(count_pairs.drop(columns='count') , how='inner')
    
    # Keep only samples above a certain number of LRs
    count_lrs = (liana_res.
                 drop(columns=score_key).
                 groupby([sample_key, 'ct_pair']).
                 count().
                 rename(columns={'interaction': 'count'}).
                 reset_index()
                 )
    count_lrs = count_lrs[count_lrs['count'] >= lrs_per_sample]
    liana_res = liana_res.merge(count_lrs.drop(columns='count') , how='inner')
    
    # convert to anndata views
    views = liana_res['ct_pair'].unique()
    views = tqdm(views, disable=not verbose)
    
    lr_adatas = {}    
    for view in views:
        lrs_per_ct = liana_res[liana_res['ct_pair']==view]
        lrs_wide = lrs_per_ct.pivot(index='interaction', 
                                    columns=sample_key,
                                    values=score_key)
    
        lrs_wide.index = view + var_separator + lrs_wide.index
        lrs_wide = lrs_wide.replace(np.nan, lr_fill)
        
        if lrs_wide.shape[0] >= lrs_per_view: # check if enough LRs
            temp = _dataframe_to_anndata(lrs_wide)
            
            # keep only variables with variance > min_variance
            temp = temp[:, np.nanvar(temp.X, axis=0) > min_variance]
            
            if (temp.shape[0] >= samples_per_view): # check if enough samples
                lr_adatas[view] = temp
                
    # to mdata
    mdata = MuData(lr_adatas)
    
    # process metadata
    _process_meta(adata=adata, mdata=mdata, sample_key=sample_key, obs_keys=obs_keys)
    
    return mdata
        

def _dataframe_to_anndata(df):
    obs = pd.DataFrame(index=df.columns)
    var = pd.DataFrame(index=df.index)
    X = np.array(df.values).T
    
    return AnnData(X=X, obs=obs, var=var, dtype=np.float32)



def get_variable_loadings(mdata,
                          idx,
                          varm_key = 'LFs',
                          view_separator = None,
                          variable_separator = None,
                          pair_separator = None,
                          var_names = ['ligand_complex', 'receptor_complex'],
                          pair_names = ['source', 'target'],
                          drop_columns = True
                          ):
    """
    A helper function to extract variable loadings from a MuData object.
    
    Parameters
    ----------
    
    mdata: :class:`~mudata.MuData`
        MuData object
    idx: int
        Index of the variable to extract. Pass index 0 to extract the first Factor.
    varm_key: str
        Key to use when extracting variable loadings from `mdata.varm`
    view_separator: str
        Separator to use when splitting view:variable names into view and variable
    variable_separator: str
        Separator to use when splitting variable names into `var_names` ('ligand_complex' and 'receptor_complex' by default)
    pair_separator: str
        Separator to use when splitting view names into `pair_names` ('source' and 'target' by default)
    drop_columns: bool
        If True, drop the `view:variable` column
        
    Returns
    -------
    Returns a pandas DataFrame with the variable loadings for the specified index.
    
    """
    
    df = sc.get.var_df(mdata, varm_keys=[(varm_key, idx)])
    df.index.name = None
    df = df.reset_index().rename(columns={'index':'view:variable'})
    
    if view_separator is not None:
        df[['view', 'variable']] = df['view:variable'].str.split(view_separator, 1, expand=True)
        
        if drop_columns:
            df.drop(columns='view:variable', inplace=True)
    
    if variable_separator is not None:
        df[var_names] = df['variable'].str.split(variable_separator, 1, expand=True)
        
        if drop_columns:
            df.drop(columns='variable', inplace=True)
        
    
    if pair_separator is not None:
        df[pair_names] = df['view'].str.split(pair_separator, 1, expand=True)
        
        if drop_columns:
            df.drop(columns='view', inplace=True)
    
    df = df.rename(columns={"LFs-{0}".format(0):'loadings'})
    
    # re-order to absolute values
    df = (df.reindex(df['loadings'].abs().sort_values(ascending=False).index))
    
    return df


def get_factor_scores(mdata, obsm_key='X_mofa'):
    """
    A helper function to extract factor scores from a MuData object.
    
    Parameters
    ----------
    mdata: :class:`~mudata.MuData`
        MuData object
    obsm_key: str
        Key to use when extracting factor scores from `mdata.obsm`
        
    Returns
    -------
    Returns a pandas DataFrame with the factor scores.
    
    """
    
    if obsm_key not in mdata.obsm.keys():
        raise ValueError(f'{obsm_key} not found in mdata.obsm')
    
    df = pd.DataFrame(mdata.obsm['X_mofa'], index=mdata.obs.index)
    
    df.columns = ['Factor_{0}'.format(x + 1) for x in range(df.shape[1])]
    df = df.reset_index()
    
    # join with metadata
    df = df.merge(mdata.obs.reset_index())
    
    return df


def _process_meta(adata, mdata, sample_key, obs_keys):
    if obs_keys is not None:
        metadata = adata.obs[[sample_key, *obs_keys]].drop_duplicates()
        
        sample_n = adata.obs[sample_key].nunique()
        if metadata.shape[0] != sample_n:
            raise ValueError('`obs_keys` must be unique per sample in `adata.obs`')
        
        mdata.obs.index.name = None
        mdata.obs = (mdata.obs.
                     reset_index().
                     rename(columns={"index":sample_key}).
                     merge(metadata).
                     set_index(sample_key)
                     )


from liana.method._liana_pipe import liana_pipe, _expm1_base, _calc_log2fc

import pathlib
from scanpy.datasets import pbmc68k_reduced
from pandas.testing import assert_frame_equal
from pandas import read_csv
import numpy as np

test_path = pathlib.Path(__file__).parent
adata = pbmc68k_reduced()

groupby = 'bulk_labels'
de_method = 't-test'
resource_name = 'consensus'
expr_prop = 0.1
min_cells = 5
complex_policy = 'min'
key_cols = ['source', 'target', 'ligand_complex', 'receptor_complex']
verbose = False
base = 2.718281828459045
resource = None
use_raw = True
layer = None
n_perms = 5
seed = 1337


# Test ALL Default parameters
def test_liana_pipe_defaults():
    all_defaults = liana_pipe(adata=adata,
                              groupby=groupby,
                              resource_name=resource_name,
                              expr_prop=expr_prop,
                              min_cells=min_cells,
                              de_method=de_method,
                              base=base,
                              n_perms=n_perms,
                              seed=seed,
                              verbose=verbose,
                              _key_cols=key_cols,
                              supp_columns=[],
                              resource=resource,
                              use_raw=use_raw,
                              layer=layer
                              )

    assert 1288 == all_defaults.shape[0]
    assert 21 == all_defaults.shape[1]
    assert 'prop_min' in all_defaults.columns

    exp_defaults = read_csv(test_path.joinpath("data/all_defaults.csv"), index_col=0)
    exp_defaults.index = all_defaults.index
    assert_frame_equal(all_defaults, exp_defaults, check_dtype=False,
                       check_exact=False, check_less_precise=True, check_index_type=False)


# Test NOT Default parameters
def test_liana_pipe_not_defaults():
    not_defaults = liana_pipe(adata=adata,
                              groupby=groupby,
                              resource_name=resource_name,
                              expr_prop=0.2,
                              min_cells=min_cells,
                              de_method='wilcoxon',
                              base=base,
                              n_perms=n_perms,
                              seed=seed,
                              verbose=verbose,
                              _key_cols=key_cols,
                              supp_columns=['ligand_pvals', 'receptor_pvals'],
                              resource=resource,
                              use_raw=use_raw,
                              layer=layer,
                              return_all_lrs=True
                              )

    assert 4200 == not_defaults.shape[0]
    assert 24 == not_defaults.shape[1]
    assert all(np.isin(['lrs_to_keep'], not_defaults.columns))
    assert all(np.isin(['ligand_pvals', 'receptor_pvals'], not_defaults.columns))

    exp_defaults = read_csv(test_path.joinpath("data/not_defaults.csv"), index_col=0)
    exp_defaults.index = not_defaults.index
    assert_frame_equal(not_defaults, exp_defaults, check_dtype=False,
                       check_index_type=False, check_exact=False, check_less_precise=True)


def test_expm1_fun():
    expm1_mat = _expm1_base(base, adata.raw.X.data)
    np.testing.assert_almost_equal(np.sum(expm1_mat), 1057526.4, decimal=1)


def test_calc_log2fc():
    adata.layers['normcounts'] = adata.raw.X.copy()
    adata.layers['normcounts'].data = _expm1_base(base, adata.raw.X.data)
    adata.obs['label'] = adata.obs.bulk_labels
    np.testing.assert_almost_equal(np.mean(_calc_log2fc(adata, "Dendritic")), -0.123781264)



from liana.method._Method import Method, MethodMeta
from ._cellphonedb import _simple_mean


def _logfc_score(x):
    """
    Calculate 1vs Rest expression logFC

    Parameters
    ----------
    x
        DataFrame row

    Returns
    -------
    (None, 1vsRest expression logFC)

    """
    # specificity
    mean_logfc = _simple_mean(x.ligand_logfc, x.receptor_logfc)
    return None, mean_logfc


# Initialize CPDB Meta
_logfc = MethodMeta(method_name="log2FC",
                    complex_cols=['ligand_means', 'receptor_means'],
                    add_cols=['ligand_logfc', 'receptor_logfc'],
                    fun=_logfc_score,
                    magnitude=None,
                    magnitude_ascending=None,
                    specificity='lr_logfc',
                    specificity_ascending=False,
                    permute=False,
                    reference='Dimitrov, D., Türei, D., Garrido-Rodriguez, M., Burmedi, P.L., '
                              'Nagai, J.S., Boys, C., Ramirez Flores, R.O., Kim, H., Szalai, B., '
                              'Costa, I.G. and Valdeolivas, A., 2022. Comparison of methods and '
                              'resources for cell-cell communication inference from single-cell '
                              'RNA-Seq data. Nature Communications, 13(1), pp.1-13. '
                    )

# Initialize callable Method instance
logfc = Method(_SCORE=_logfc)



from ._rank_aggregate import _rank_aggregate_meta
from ._cellphonedb import cellphonedb
from ._logfc import logfc
from ._natmi import natmi
from ._connectome import connectome
from ._singlecellsignalr import singlecellsignalr
from ._geometric_mean import geometric_mean
from ._cellchat import cellchat



import pathlib

from unittest import TestCase

from pandas import read_csv
from pandas.testing import assert_frame_equal
from numpy import random

from liana.method import rank_aggregate
from liana.method.sc._rank_aggregate import AggregateClass
from liana.testing._toy_adata import get_toy_adata



test_path = pathlib.Path(__file__).parent

adata = get_toy_adata()


def test_consensus():
    assert isinstance(rank_aggregate, AggregateClass)
    assert rank_aggregate.magnitude == 'magnitude_rank'
    assert rank_aggregate.specificity == 'specificity_rank'
    assert rank_aggregate.steady == 'steady_rank'
    assert rank_aggregate.method_name == 'Rank_Aggregate'


def test_aggregate_specs():
    steady_specs = {'CellPhoneDB': ('cellphone_pvals', True),
                    'Connectome': ('scaled_weight', False),
                    'log2FC': ('lr_logfc', False),
                    'NATMI': ('spec_weight', False),
                    'SingleCellSignalR': ('lrscore', False),
                    'CellChat': ('cellchat_pvals', True),
                    }
    TestCase().assertDictEqual(rank_aggregate.steady_specs, steady_specs)

    magnitude_specs = {'CellPhoneDB': ('lr_means', False),
                       'Connectome': ('expr_prod', False),
                       'NATMI': ('expr_prod', False),
                       'SingleCellSignalR': ('lrscore', False),
                       'CellChat': ('lr_probs', False)
                       }

    TestCase().assertDictEqual(rank_aggregate.magnitude_specs, magnitude_specs)


def test_aggregate_res():
    lr_res = rank_aggregate(adata, groupby='bulk_labels', use_raw=True, n_perms=2, inplace=False)
    lr_exp = read_csv(test_path.joinpath("data/aggregate_rank_rest.csv"), index_col=0)

    assert_frame_equal(lr_res, lr_exp, check_dtype=False,
                       check_exact=False, check_less_precise=True)


def test_aggregate_all():
    rank_aggregate(adata, groupby='bulk_labels', use_raw=True, return_all_lrs=True, key_added='all_res')
    assert adata.uns['all_res'].shape == (4200, 16)


def test_aggregate_by_sample():
    
    rank_aggregate.by_sample(adata, groupby='bulk_labels', use_raw=True, return_all_lrs=True, sample_key='sample', key_added='liana_by_sample')
    lr_by_sample = adata.uns['liana_by_sample']
    
    assert 'sample' in lr_by_sample.columns
    assert lr_by_sample.shape == (10836, 17)


__version__ = '0.1.8'
__version_info__ = tuple([int(num) for num in __version__.split('.')])

from liana import method as mt, plotting as pl, resource as rs, multi as mu, funcomics as fun, testing

# done after everything has been imported (adapted from scanpy)
import sys
from scanpy._utils import annotate_doc_types

sys.modules.update({f'{__name__}.{m}': globals()[m] for m in ['mt', 'pl', 'rs', 'mu', 'fun']})
annotate_doc_types(sys.modules[__name__], 'liana')


del sys, annotate_doc_types



import numpy as np
import pandas as pd
import string


# Function to generate a liana_res sample/example
def sample_lrs(by_sample=False):
    """Create sample method output for testing metrics in this task."""
    row_num = 200
    rng = np.random.default_rng(seed=1)

    label_vector = list(string.ascii_uppercase[0:10])
    entity_vector = list(string.ascii_lowercase[0:10])

    df = pd.DataFrame(rng.random((row_num, 1)), columns=["magnitude"])
    df["source"] = rng.choice(label_vector, row_num)
    df["target"] = rng.choice(label_vector, row_num)
    df["ligand_complex"] = rng.choice(entity_vector, row_num)
    df["receptor_complex"] = rng.choice(entity_vector, row_num)
    df["specificity_rank"] = rng.random((200, 1))        

    # deduplicate
    df = df.loc[~df.duplicated()]
    
    if by_sample:
        df['sample'] = rng.choice(['A', 'B', 'C', 'D'], row_num)
        df['sample'] = df['sample'].astype('category')

    return df



from __future__ import annotations

import anndata
import numpy as np
import pandas
from tqdm import tqdm
from statsmodels.distributions.empirical_distribution import ECDF


def _get_means_perms(adata: anndata.AnnData,
                     lr_res: pandas.DataFrame,
                     n_perms: int,
                     seed: int,
                     agg_fun,
                     norm_factor: float | None,
                     verbose: bool):
    """
    Generate permutations and indices required for permutation-based methods

    Parameters
    ----------
    adata
        Annotated data matrix.
    lr_res
        Ligand-receptor stats DataFrame
    n_perms
        Number of permutations to be calculated
    seed
        Random seed for reproducibility.
    agg_fun
        function by which to aggregate the matrix, should take `axis` argument
    norm_factor
        additionally normalize the data by some factor (e.g. matrix max for CellChat)
    verbose
        Verbosity bool

    Returns
    -------
    Tuple with:
        - perms: 3D tensor with permuted averages per cluster
        - ligand_pos: Index of the ligand in the tensor
        - receptor_pos: Index of the receptor in the perms tensor
        - labels_pos: Index of cell identities in the perms tensor
    """

    # initialize rng
    rng = np.random.default_rng(seed=seed)

    if isinstance(norm_factor, np.float):
        adata.X /= norm_factor

    # define labels and dict
    labels = adata.obs.label.cat.categories
    labels_dict = {label: adata.obs.label.isin([label]) for label in labels}

    # indexes to be shuffled
    idx = np.arange(adata.X.shape[0])

    # Perm should be a cube /w dims: n_perms x idents x n_genes
    perms = np.zeros((n_perms, labels.shape[0], adata.shape[1]))

    # Assign permuted matrix
    for perm in tqdm(range(n_perms), disable=not verbose):
        perm_idx = rng.permutation(idx)
        perm_mat = adata.X[perm_idx]
        # populate matrix /w permuted means
        for cind in range(labels.shape[0]):
            perms[perm, cind] = agg_fun(perm_mat[labels_dict[labels[cind]]], axis=0)

    # Get indexes for each gene and label in the permutations
    ligand_pos = {entity: np.where(adata.var_names == entity)[0][0] for entity
                  in lr_res['ligand']}
    receptor_pos = {entity: np.where(adata.var_names == entity)[0][0] for entity
                    in lr_res['receptor']}
    labels_pos = {labels[pos]: pos for pos in range(labels.shape[0])}

    return perms, ligand_pos, receptor_pos, labels_pos


def _get_lr_pvals(x, perms, ligand_pos, receptor_pos, labels_pos, agg_fun,
                  ligand_col='ligand_means', receptor_col='receptor_means'):
    """
    Calculate Permutation means and p-values

    Parameters
    ----------
    x
        DataFrame row
    perms
        3D tensor with permuted averages per cluster
    ligand_pos
        Index of the ligand in the tensor
    receptor_pos
        Index of the receptor in the perms tensor
    labels_pos
        Index of cell identities in the perms tensor
    agg_fun
        function to aggregate the ligand and receptor

    Returns
    -------
    A tuple with lr_score (aggregated according to `agg_fun`) and ECDF p-value for x

    """
    # actual lr_score
    lr_score = agg_fun(x[ligand_col], x[receptor_col])

    if lr_score == 0:
        return 0, 1

    # Permutations lr mean
    ligand_perm_means = perms[:, labels_pos[x.source], ligand_pos[x.ligand]]
    receptor_perm_means = perms[:, labels_pos[x.target], receptor_pos[x.receptor]]
    lr_perm_score = agg_fun(ligand_perm_means, receptor_perm_means)

    p_value = (1 - ECDF(lr_perm_score)(lr_score))

    return lr_score, p_value



from numpy import random

from scanpy.datasets import pbmc68k_reduced

def get_toy_adata():
    adata = pbmc68k_reduced()
    sample_key = 'sample'
    
    rng = random.default_rng(0)
    
    # create fake samples
    adata.obs[sample_key] = rng.choice(['A', 'B', 'C', 'D'], size=len(adata.obs))
    
    # group samples into conditions
    adata.obs['case'] = adata.obs[sample_key].map({'A': 'yes', 'B': 'yes', 'C': 'no', 'D': 'no'})
    
    return adata
    


import numpy as np
import pathlib
from scanpy.datasets import pbmc68k_reduced
from pandas import read_csv

from liana.method._pipe_utils._get_mean_perms import _get_means_perms
from liana.method._liana_pipe import _trimean

test_path = pathlib.Path(__file__).parent

adata = pbmc68k_reduced()
adata.X = adata.raw.X
adata.obs['label'] = adata.obs.bulk_labels

all_defaults = read_csv(test_path.joinpath("data/all_defaults.csv"), index_col=0)

perms, ligand_pos, receptor_pos, labels_pos = \
    _get_means_perms(adata=adata, lr_res=all_defaults,
                     norm_factor=None, agg_fun=np.mean,
                     n_perms=100, seed=1337, verbose=False)


def test_perms():
    assert perms.shape == (100, 10, 765)

    desired = np.array([45615.15418553, 45737.95729483, 45575.47318892, 45559.41832494,
                        45542.32316456, 45593.50440302, 45591.03955124, 45561.79108855,
                        45698.86540851, 45543.95444739])
    expected = np.sum(np.sum(perms, axis=0), axis=1)

    assert np.testing.assert_almost_equal(desired, expected, decimal=3) is None


def test_positions():
    assert ligand_pos['MIF'] == 740
    assert receptor_pos['CD4'] == 465
    assert labels_pos['Dendritic'] == 9


def test_cellchat_perms():
    mat_max = np.max(adata.X.data)

    perms, _, _, _ = \
        _get_means_perms(adata=adata, lr_res=all_defaults,
                         norm_factor=mat_max, agg_fun=_trimean,
                         n_perms=100, seed=1337, verbose=False)

    assert perms.shape == (100, 10, 765)

    desired = np.array([33840.83, 36332.442, 34569.577, 33819.275,
                        33809.956, 33785.234, 33844.524, 34986.043,
                        34304.404, 33644.323])
    expected = np.sum(np.sum(perms, axis=0), axis=1)

    assert np.testing.assert_almost_equal(desired, expected, decimal=3) is None



from __future__ import annotations

from ._liana_pipe import liana_pipe

from anndata import AnnData
from pandas import DataFrame, concat
from typing import Optional
from tqdm import tqdm
import weakref


class MethodMeta:
    """
    A Class used to store Method Metadata
    """
    
    # initiate a list to store weak references to all instances
    instances = [] ## TODO separate instances for each child class
    
    def __init__(self,
                 method_name: str,
                 complex_cols: list,
                 add_cols: list,
                 fun,
                 magnitude: str | None,
                 magnitude_ascending: bool | None,
                 specificity: str | None,
                 specificity_ascending: bool | None,
                 permute: bool,
                 reference: str
                 ):
        """
        Parameters
        ----------
        method_name
            Name of the Method
        complex_cols
            Columns relevant for protein complexes
        add_cols
            Additional columns required by the method
        fun
            Interaction Scoring function
        magnitude
            Name of the `magnitude` Score (None if not present)
        magnitude_ascending
            Whether to rank `magnitude` in ascending manner (None if not relevant)
        specificity
            Name of the `specificity` Score if Present (None if not present)
        specificity_ascending
            Whether to rank `magnitude` in ascending manner  (None if not relevant)
        permute
            Whether it requires permutations
        reference
            Publication reference in Harvard style
        """
        self.__class__.instances.append(weakref.proxy(self))
        self.method_name = method_name
        self.complex_cols = complex_cols
        self.add_cols = add_cols
        self.fun = fun
        self.magnitude = magnitude
        self.magnitude_ascending = magnitude_ascending
        self.specificity = specificity
        self.specificity_ascending = specificity_ascending
        self.permute = permute
        self.reference = reference

    # describe self
    def describe(self):
        """Briefly described the method"""
        print(
            f"{self.method_name} uses `{self.magnitude}` and `{self.specificity}`"
            f" as measures of expression strength and interaction specificity, respectively"
        )

    def reference(self):
        """Prints out reference in Harvard format"""
        print(self.reference)

    def get_meta(self):
        """Returns method metadata as pandas row"""
        meta = DataFrame([{"Method Name": self.method_name,
                           "Magnitude Score": self.magnitude,
                           "Specificity Score": self.specificity,
                           "Reference": self.reference
                           }])
        return meta
    
    def by_sample(self, adata, sample_key, key_added='liana_res', inplace=True, verbose=False, **kwargs):
        """
        Run a method by sample.
        
        Parameters
        ----------
            adata 
                AnnData object to run the method on
            sample_key
                key in `adata.obs` to use for grouping by sample/context
            key_added
                key to store the results in `adata.uns` if `inplace` is True
            inplace
                whether to store the results in `adata.uns['liana_res']` or return a dataframe
            verbose
                Possible values: False, True, 'full', where 'full' will print the results for each sample,
                and True will only print the sample progress bar. Default is False.
            **kwargs
                keyword arguments to pass to the method
        
        Returns
        -------
        A pandas DataFrame with the results and a column sample if inplace is False, else None
        
        """
        
        if sample_key not in adata.obs:
            raise ValueError(f"{sample_key} was not found in `adata.obs`.")
        
        if adata.obs[sample_key].dtype.name != "category":
            (f"`{sample_key}` was assigned as a categorical.")
            adata.obs[sample_key] = adata.obs[sample_key].astype("category")
            
        if verbose == 'full':
            verbose = True
            full_verbose = True
        else:
            full_verbose = False
            
        samples = adata.obs[sample_key].cat.categories 
            
        adata.uns[key_added] = {}
        
        progress_bar = tqdm(samples, disable=not verbose)
        for sample in (progress_bar):
            if verbose:
                progress_bar.set_description(f"Now running: {sample}")

            temp = adata[adata.obs[sample_key]==sample].copy()

            sample_res = self.__call__(temp, inplace=False, verbose=full_verbose, **kwargs)

            adata.uns[key_added][sample] = sample_res

        liana_res = concat(adata.uns[key_added]).reset_index(level=1, drop=True).reset_index()
        liana_res = liana_res.rename({"index":sample_key}, axis=1)
        
        if inplace:
            adata.uns[key_added] = liana_res
        return None if inplace else liana_res


class Method(MethodMeta):
    """
    liana's Method Class
    """
    def __init__(self, _SCORE):
        super().__init__(method_name=_SCORE.method_name,
                         complex_cols=_SCORE.complex_cols,
                         add_cols=_SCORE.add_cols,
                         fun=_SCORE.fun,
                         magnitude=_SCORE.magnitude,
                         magnitude_ascending=_SCORE.magnitude_ascending,
                         specificity=_SCORE.specificity,
                         specificity_ascending=_SCORE.specificity_ascending,
                         permute=_SCORE.permute,
                         reference=_SCORE.reference
                         )
        self._SCORE = _SCORE

    def __call__(self,
                 adata: AnnData,
                 groupby: str,
                 resource_name: str = 'consensus',
                 expr_prop: float = 0.1,
                 min_cells: int = 5,
                 base: float = 2.718281828459045,
                 supp_columns: list = None,
                 return_all_lrs: bool = False,
                 key_added: str = 'liana_res',
                 use_raw: Optional[bool] = True,
                 layer: Optional[str] = None,
                 de_method='t-test',
                 verbose: Optional[bool] = False,
                 n_perms: int = 1000,
                 seed: int = 1337,
                 resource: Optional[DataFrame] = None,
                 inplace=True):
        """
        Parameters
        ----------
        adata
            Annotated data object.
        groupby
            The key of the observations grouping to consider.
        resource_name
            Name of the resource to be loaded and use for ligand-receptor inference.
        expr_prop
            Minimum expression proportion for the ligands/receptors (and their subunits) in the
            corresponding cell identities. Set to `0`, to return unfiltered results.
        min_cells
            Minimum cells per cell identity (`groupby`) to be considered for downstream analysis
        base
            Exponent base used to reverse the log-transformation of matrix. Note that this is
            relevant only for the `logfc` method.
        supp_columns
            Any additional columns to be added from any of the methods implemented in
            liana, or any of the columns returned by `scanpy.tl.rank_genes_groups`, each
            starting with ligand_* or receptor_*. For example, `['ligand_pvals', 'receptor_pvals']`.
            `None` by default.
        return_all_lrs
            Bool whether to return all LRs, or only those that surpass the `expr_prop`
            threshold. Those interactions that do not pass the `expr_prop` threshold will
            be assigned to the *worst* score of the ones that do. `False` by default.
        key_added
            Key under which the results will be stored in `adata.uns` if `inplace` is True.
        use_raw
            Use raw attribute of adata if present.
        layer
            Layer in anndata.AnnData.layers to use. If None, use anndata.AnnData.X.
        de_method
            Differential expression method. `scanpy.tl.rank_genes_groups` is used to rank genes
            according to 1vsRest. The default method is 't-test'. Only relevant if p-values
            are included in `supp_cols`
        verbose
            Verbosity flag
        n_perms
            Number of permutations for the permutation test. Note that this is relevant
            only for permutation-based methods - e.g. `CellPhoneDB`
        seed
            Random seed for reproducibility.
        resource
            Parameter to enable external resources to be passed. Expects a pandas dataframe
            with [`ligand`, `receptor`] columns. None by default. If provided will overrule
            the resource requested via `resource_name`
        inplace
            If true return `DataFrame` with results, else assign to `.uns`.

        Returns
        -------
            If ``inplace = False``, returns a `DataFrame` with ligand-receptor results
            Otherwise, modifies the ``adata`` object with the following key:
            - :attr:`anndata.AnnData.uns` ``[`key_added`]`` with the aforementioned DataFrame
        """
        if supp_columns is None:
            supp_columns = []

        liana_res = liana_pipe(adata=adata,
                               groupby=groupby,
                               resource_name=resource_name,
                               resource=resource,
                               expr_prop=expr_prop,
                               min_cells=min_cells,
                               supp_columns=supp_columns,
                               return_all_lrs=return_all_lrs,
                               base=base,
                               de_method=de_method,
                               verbose=verbose,
                               _score=self._SCORE,
                               n_perms=n_perms,
                               seed=seed,
                               use_raw=use_raw,
                               layer=layer,
                               )
        if inplace:
            adata.uns[key_added] = liana_res
        return None if inplace else liana_res
        

def _show_methods(methods):
    return concat([method.get_meta() for method in methods])



import numpy as np


def _assign_entity_weights(resource, net, entity_key='receptor', source='source', target='target', weight='weight'):
    # only keep relevant columns
    net = net[[source, target, weight]]
    
    # process ligand-receptor resource
    # assign receptor complex as entity
    entity_resource = resource[[entity_key]].drop_duplicates().set_index(entity_key)
    entity_resource['subunit'] =  entity_resource.index
    # explode complexes, keeping the complex as a key
    entity_resource = entity_resource.apply(lambda x: x.str.split('_')).explode(['subunit'])
    
    # join weights to subunits
    entity_resource = entity_resource.reset_index()
    entity_resource = entity_resource.merge(net, left_on='subunit', right_on=target)
    
    # check for sign and set consistency
    # count expected subunits separated by _
    entity_resource = entity_resource.assign(subunit_expected = entity_resource[entity_key].str.count('_')+1)
    # count subunits by receptor complex & source
    entity_resource['subunit_count'] = entity_resource.groupby([source, entity_key])[[weight]].transform('count')
    # check if all subunits are present
    entity_resource = entity_resource.assign(subunit_complete = lambda x: x['subunit_expected'] == x['subunit_count'])
    # assign flag to sign-coherent subunits
    entity_resource['sing_coherent'] = entity_resource.groupby([source, entity_key])[[weight]].transform(lambda x: np.all(x > 0) | np.all(x < 0))
    
    # keep only relevant targets
    entity_resource = entity_resource[entity_resource['subunit_complete']] # keep only complete complexes
    entity_resource = entity_resource[entity_resource['sing_coherent']] # keep only sign-coherent complexes
    
    # get mean weight per complex & source
    entity_resource = entity_resource.groupby([source, entity_key])[[weight]].mean().reset_index()
    
    return entity_resource


# function that returns the mean only if all values are sign coherent
def _sign_coherent_mean(x):
    if np.all(x > 0) | np.all(x < 0):
        return np.mean(x)
    else:
        return np.nan
    
    
    
def generate_lr_geneset(resource,
                        net, 
                        ligand='ligand',
                        receptor='receptor',
                        lr_separator='&',
                        source='source', 
                        weight='weight'):
    """
    Generate a ligand-receptor gene set from a resource and a network.
    
    Specifically, it works with weighted bipartite networks, where the weight represents the importance of the genes 
    to a given geneset. The function will assign a weight to each ligand-receptor interaction, based on the mean.
    It does so by first assigning a weight to each ligand-receptor subunit, checking for sign coherence and completeness
    of the ligand-receptor complex.
    
    Parameters
    ----------
    resource : pandas.DataFrame
        Ligand-receptor resource.
    net : pandas.DataFrame
        Prior knowledge network in bipartite or decoupler format.
    ligand : str, optional
        Name of the ligand column in the resource, by default 'ligand'
    receptor : str, optional
        Name of the receptor column in the resource, by default 'receptor'
    lr_separator : str, optional
        Separator to use when joining ligand and receptor, by default '&'
    source : str, optional
        Name of the source column in the network, by default 'source'
    weight : str, optional
    
    Returns
    -------
    Returns ligand-receptor geneset resource as a pandas.DataFrame with the following columns:
    - interaction: ligand-receptor interaction
    - weight: mean weight of the interaction
    - source: source of the interaction
    """
    
    # supp keys
    ligand_weight = ligand + '_' + weight
    receptor_weight = receptor + '_' + weight
    ligand_source = ligand + '_' + source
    receptor_source = receptor + '_' + source
    
    
    # assign weights to each entity
    ligand_weights = _assign_entity_weights(resource, net, entity_key=ligand)
    ligand_weights.rename(columns={weight: ligand_weight, source:ligand_source}, inplace=True)
    receptor_weights = _assign_entity_weights(resource, net, entity_key=receptor)
    receptor_weights.rename(columns={weight: receptor_weight, source: receptor_source}, inplace=True)
    
    # join weights to the the ligand-receptor resource
    resource = resource.merge(ligand_weights, on=ligand, how='inner')
    resource = resource.merge(receptor_weights, on=receptor, how='inner')
    
    # keep only coherent ligand and receptor sources
    resource = resource[resource[ligand_source] == resource[receptor_source]]
    # mean of sign-coherent ligand-receptor weights
    resource.loc[:, weight] = resource.apply(lambda x: _sign_coherent_mean(np.array([x[ligand_weight], x[receptor_weight]])), axis=1)
    
    # unite ligand-receptor columns
    resource = resource.assign(interaction = lambda x: x[ligand] + lr_separator + x[receptor])
    
    # keep only relevant columns
    resource = resource[[ligand_source, 'interaction', weight]].rename(columns={ligand_source: source})
    
    # drop nan weights
    resource = resource.dropna()
    
    return resource



"""
Preprocessing functions.
Functions to preprocess the anndata object prior to running any method.
"""

import numpy as np
from anndata import AnnData
from typing import Optional
from pandas import DataFrame, Index
import scanpy as sc
from scipy.sparse import csr_matrix, isspmatrix_csr


def assert_covered(
        subset,
        superset,
        subset_name: str = "resource",
        superset_name: str = "var_names",
        prop_missing_allowed: float = 0.98,
        verbose: bool = False) -> None:
    """
    Assert if elements are covered at a decent proportion

    Parameters
    ----------
    subset
        Subset of elements
    superset
        The superset of elements
    subset_name
        Name of the subset
    superset_name
        Name of the superset
    prop_missing_allowed
        Allowed proportion of missing/mismatched elements in the subset
    verbose
        Verbosity flag

    Returns
    -------
    None
    """

    subset = np.asarray(subset)
    is_missing = ~np.isin(subset, superset)
    prop_missing = np.sum(is_missing) / len(subset)
    x_missing = ", ".join([x for x in subset[is_missing]])

    if prop_missing > prop_missing_allowed:
        msg = (
            f"Please check if appropriate organism/ID type was provided! "
            f"Allowed proportion ({prop_missing_allowed}) of missing "
            f"{subset_name} elements exceeded ({prop_missing:.2f}). "
            f"Too few features from the resource were found in the data."
        )
        raise ValueError(msg + f" [{x_missing}] missing from {superset_name}")
    if verbose & (prop_missing > 0):
        print(f"{prop_missing:.2f} of entities in the resource are missing from the data.")


def prep_check_adata(adata: AnnData,
                     groupby: str,
                     min_cells: int,
                     use_raw: Optional[bool] = False,
                     layer: Optional[str] = None,
                     verbose: Optional[bool] = False) -> AnnData:
    """
    Check if the anndata object is in the correct format and preprocess

    Parameters
    ----------
    adata
        Un-formatted Anndata.
    groupby
        column to groupby
    min_cells
        minimum cells per cell identity
    use_raw
        Use raw attribute of adata if present.
    layer
        Indicate whether to use any layer.
    verbose
        Verbosity flag.

    Returns
    -------
    Anndata object to be used downstream
    """
    # simplify adata

    X = _choose_mtx_rep(adata=adata, use_raw=use_raw,
                        layer=layer, verbose=verbose)

    if use_raw & (layer is None):
        var = adata.raw.var.copy()
    else:
        var = adata.var.copy()

    adata = sc.AnnData(X=X,
                       obs=adata.obs.copy(),
                       dtype="float32",
                       var=var
                       )

    # convert to sparse csr matrix
    if not isspmatrix_csr(adata.X):
        if verbose:
            print("Converting mat to CSR format")
        adata.X = csr_matrix(adata.X)

    # Check for empty features
    msk_features = np.sum(adata.X, axis=0).A1 == 0
    n_empty_features = np.sum(msk_features)
    if n_empty_features > 0:
        if verbose:
            print("{0} features of mat are empty, they will be removed.".format(
                n_empty_features))
        adata = adata[:, ~msk_features]

    # Check for empty samples
    msk_samples = np.sum(adata.X, axis=1).A1 == 0
    n_empty_samples = np.sum(msk_samples)
    if n_empty_samples > 0:
        if verbose:
            print("{0} samples of mat are empty, they will be removed.".format(
                n_empty_samples))
        adata = adata[~msk_samples, :]

    # Check if log-norm
    _sum = np.sum(adata.X.data[0:100])
    if _sum == np.floor(_sum):
        if verbose:
            print("Make sure that normalized counts are passed!")

    # Check for non-finite values
    if np.any(~np.isfinite(adata.X.data)):
        raise ValueError(
            """mat contains non finite values (nan or inf), please set them 
            to 0 or remove them.""")

    # Define idents col name
    if groupby not in adata.obs.columns:
        raise AssertionError(f"`{groupby}` not found in `adata.obs.columns`.")
    adata.obs.loc[:, 'label'] = adata.obs[groupby]

    # Re-order adata vars alphabetically
    adata = adata[:, np.sort(adata.var_names)]

    # Remove any cell types below X number of cells per cell type
    count_cells = adata.obs.groupby(groupby)[groupby].size().reset_index(name='count').copy()
    count_cells['keep'] = count_cells['count'] >= min_cells

    if not all(count_cells.keep):
        lowly_abundant_idents = list(count_cells[~count_cells.keep][groupby])
        # remove lowly abundant identities
        msk = ~np.isin(adata.obs[[groupby]], lowly_abundant_idents)
        adata = adata[msk]
        if verbose:
            print("The following cell identities were excluded: {0}".format(
                ", ".join(lowly_abundant_idents)))

    # Remove underscores from gene names
    adata.var_names = format_vars(adata.var_names)

    return adata


# Helper function to replace a substring in string and append to list
def _append_replace(x: str, l: list):
    l.append(x)
    return x.replace('_', '')


# format variable names
def format_vars(var_names, verbose=False) -> list:
    """
    Format Variable names
    
    Parameters
    ----------
    var_names
        list of variable names (e.g. adata.var_names)
    verbose
        Verbosity flag

    Returns
    -------
        Formatted Variable names list

    """
    changed = []
    var_names = [_append_replace(x, changed) if ('_' in x) else x for x in var_names]
    changed = ' ,'.join(changed)
    if verbose & (len(changed) > 0):
        print(f"Replace underscores (_) with blank in {changed}", )
    return var_names


def filter_resource(resource: DataFrame, var_names: Index) -> DataFrame:
    """
    Filter interactions for which vars are not present.

    Note that here I remove any interaction that /w genes that are not found
    in the dataset. Note that this is not necessarily the case in liana-r.
    There, I assign the expression of those with missing subunits to 0, while
    those without any subunit present are implicitly filtered.

    Parameters
    ---------
    resource
        Resource with 'ligand' and 'receptor' columns
    var_names
        Relevant variables - i.e. the variables to be used downstream

    Returns
    ------
    A filtered resource dataframe
    """
    # Remove those without any subunit
    resource = resource[(np.isin(resource.ligand, var_names)) &
                        (np.isin(resource.receptor, var_names))]

    # Only keep interactions /w complexes for which all subunits are present
    missing_comps = resource[['_' in x for x in resource['interaction']]].copy()
    missing_comps['all_units'] = \
        missing_comps['ligand_complex'] + '_' + missing_comps[
            'receptor_complex']

    # Get those not with all subunits
    missing_comps = missing_comps[np.logical_not(
        [all([x in var_names for x in entity.split('_')])
         for entity in missing_comps.all_units]
    )]
    # Filter them
    return resource[~resource.interaction.isin(missing_comps.interaction)]


def _choose_mtx_rep(adata, use_raw=False, layer=None, verbose=False) -> csr_matrix:
    """
    Choose matrix (adapted from scanpy)
    
    Parameters
    ----------
    adata
        Unformatted Anndata.
    use_raw
        Use raw attribute of adata if present.
    layer
        Indicate whether to use any layer.
    
    Returns
    -------
        The matrix to be used by liana-py.
    """

    is_layer = layer is not None
    if is_layer & use_raw:
        raise ValueError("Cannot specify `layer` and have `use_raw=True`.")
    if is_layer:
        if verbose:
            print(f"Using the `{layer}` layer!")
        return adata.layers[layer]
    elif use_raw:
        if adata.raw is None:
            raise ValueError("`.raw` is not initialized!")
        if verbose:
            print("Using `.raw`!")
        return adata.raw.X
    else:
        if verbose:
            print("Using `.X`!")
        return adata.X



"""
Utility functions to query OmniPath.
Functions to retrieve resources from the meta-database OmniPath.
"""

from json import loads
import pandas as pd


def check_if_omnipath():
    """
    Function to check if available and return OmniPath
    
    Returns
    -------
    OmniPath package

    """
    try:
        import omnipath as op
    except Exception:
        raise ImportError('omnipath is not installed. Please install it with: pip install omnipath')
    return op


# Function to explode complexes (decomplexify Resource)
def explode_complexes(resource: pd.DataFrame,
                      SOURCE='ligand',
                      TARGET='receptor') -> pd.DataFrame:
    """
    Function to explode ligand-receptor complexes

    Parameters
    ----------
    resource
        Ligand-receptor resource
    SOURCE
        Name of the source (typically ligand) column
    TARGET
        Name of the target (typically receptor) column

    Returns
    -------
    A resource with exploded complexes

    """
    resource['interaction'] = resource[SOURCE] + '|' + resource[TARGET]
    resource = (resource.set_index('interaction')
                .apply(lambda x: x.str.split('_'))
                .explode([TARGET])
                .explode(SOURCE)
                .reset_index()
                )
    resource[[f'{SOURCE}_complex', f'{TARGET}_complex']] = resource[
        'interaction'].str.split('|', expand=True)

    return resource

# """Functions to obtain additional OmniPath resources"""
# def obtain_extra_resource(databases,
#                           blocklist,
#                           allowlist
#                           ):
#     omnipath = check_if_omnipath()
#
#     # Obtain resource
#     add = omnipath.interactions.PostTranslational.get(databases=databases,
#                                                       genesymbols=True,
#                                                       entity_types=['protein', 'complex'],
#                                                       fields={"extra_attrs"}
#                                                       )
#     add = add[~add[['source', 'target']].duplicated()]
#
#     block_keys = blocklist.keys()
#     allow_keys = allowlist.keys()  # union of relevant checks
#     union_keys = block_keys ^ allow_keys
#
#     # explode relevant attributes and join them to database
#     explode_attrs = add['extra_attrs'].apply(_json_intersect_serialize, union_keys=union_keys)
#     add = pd.concat([add, explode_attrs], axis=1).drop('extra_attrs', axis=1)
#
#     # Convert blocklist to mask & remove unwanted rows
#     for k in block_keys:
#         add[k + '_msk'] = [
#             any([block in att for block in blocklist[k]]) if type(att) is not float else True
#             for att in add[k]]
#         # iter
#         add = add[~add[k + '_msk']]
#
#     # Convert allowlist to mask & keep only relevant rows
#     for k in allow_keys:
#         add[k + '_msk'] = [
#             any([allow in att for allow in allowlist[k]]) if type(att) is not float else False
#             for att in add[k]]
#         add = add[add[k + '_msk']]
#
#     return add
#
#
# # Function to format extra_attributes
# def _json_intersect_serialize(att, union_keys):
#     att = loads(att)
#     att = {k: att[k] for k in union_keys if k in att.keys()}
#     return pd.Series(att)



from ..method import get_method_scores

import numpy as np
from types import ModuleType

from ._common import _process_scores

def _check_if_tensorc2c() -> ModuleType:

    try:
        import cell2cell as c2c

    except Exception:

        raise ImportError(
            'cell2cell is not installed. Please install it with: '
            'pip install cell2cell'
        )

    return c2c


def to_tensor_c2c(adata=None,
                  liana_res=None,
                  sample_key=None,
                  source_key='source',
                  target_key='target',
                  ligand_key='ligand_complex',
                  receptor_key='receptor_complex',
                  score_key=None,
                  uns_key = 'liana_res',
                  non_expressed_fill=None,
                  inverse_fun = lambda x: 1 - x,
                  non_negative = True,
                  return_dict=False,
                  **kwargs
                  ):
    """
    Function to convert a LIANA result to a tensor for cell2cell analysis.
    
    Parameters
    ----------
    
    adata : :class:`~anndata.AnnData`
        Annotated data matrix.
    liana_res : :class:`~pandas.DataFrame`
        LIANA result.
    sample_key : `str`, optional (default: None)
        Column name of the sample key in `liana_res`.
    source_key : `str`, optional (default: 'source')
        Column name of the sender/source cell types in `liana_res`.
    target_key : `str`, optional (default: 'target')
        Column name of the receiver/target cell types in `liana_res`.
    ligand_key : `str`, optional (default: 'ligand_complex')
        Column name of the ligand in `liana_res`.
    receptor_key : `str`, optional (default: 'receptor_complex')
        Column name of the receptor in `liana_res`.
    score_key : `str`, optional (default: None)
        Column name of the score in `liana_res`. If None, the score is inferred from the method.
    inverse_fun : `function`, optional (default: lambda x: 1 - x)
        Function to inverse the score. For example, if the score is in ascending order or probability,
        the inverse function should be 1 - probability. This is handled automatically for the scores in liana.
    non_expressed_fill : `float`, optional (default: None)
        Value to fill for non-expressed ligand-receptor pairs.
    non_negative : `bool`, optional (default: True)
        Whether to make the tensor non-negative.
    return_dict : `bool`, optional (default: False)
        Whether to return a dictionary of tensors.
    **kwargs : keyword arguments to pass to Tensor-cell2cell's `cell2cell.tensor.external_scores.dataframes_to_tensor` function.
        
    Returns
    -------
    Returns a tensor of shape (n_samples, n_senders, n_receivers, n_interactions) or a dictionary of tensors if `return_dict` is True.
        
    
    """
    
    # check if cell2cell is installed
    c2c = _check_if_tensorc2c()
    
    if (liana_res is None) & (adata is None):
        raise AttributeError('Ambiguous! One of `liana_res` or `adata` should be provided.')
    if adata is not None:
        assert uns_key in adata.uns_keys()
        liana_res = adata.uns[uns_key].copy()
    if liana_res is not None:
        liana_res = liana_res.copy()
    if (liana_res is None) & (adata is None):
        raise ValueError('`liana_res` or `adata` must be provided!')
    
    keys = np.array([sample_key, source_key, target_key, ligand_key, receptor_key])
    missing_keys = keys[[ key not in liana_res.columns for key in keys]]
    
    if any(missing_keys):
        raise ValueError(f'`{missing_keys}` not found in `adata.uns[{uns_key}]`! Please check your input.')
    
    # remove unneeded columns
    keys = [sample_key, source_key, target_key, ligand_key, receptor_key, score_key]
    keys = keys + ['lrs_to_keep'] if 'lrs_to_keep' in liana_res.columns else keys
    liana_res = liana_res[keys]
    
    
    # check for duplicates
    if liana_res[[sample_key, source_key, target_key, ligand_key, receptor_key]].duplicated().any():
        raise ValueError("Duplicate rows found in the input data")

    liana_res = _process_scores(liana_res, score_key, inverse_fun)

    # set negative to 0
    if non_negative:
        liana_res[score_key] = liana_res[score_key].clip(lower=0)

    # set non-expressed to 0 (if lrs_to_keep column is present)
    if ('lrs_to_keep' in liana_res.columns) & (non_expressed_fill is not None):
        liana_res.loc[~liana_res['lrs_to_keep'], score_key] = non_expressed_fill

    # split into dictionary by sample
    liana_res = {sample:df for sample, df in liana_res.groupby(sample_key)}
    
    if return_dict:
        return liana_res
    
    tensor = c2c.tensor.dataframes_to_tensor(liana_res,
                                             sender_col=source_key,
                                             receiver_col=target_key,
                                             ligand_col=ligand_key,
                                             receptor_col=receptor_key,
                                             score_col=score_key,
                                             **kwargs)

    return tensor



import numpy as np

from ..method import get_method_scores

def _process_scores(liana_res, score_key, inverse_fun):
    
    df = liana_res.copy()
    scores = get_method_scores()
    
    if not np.isin(score_key, list(scores.keys())).any():
        raise ValueError(f"Score column {score_key} not found in liana's method scores. ")
    
    # reverse if ascending order
    ascending_order = scores[score_key]
    if(ascending_order):
        df[score_key] = inverse_fun(df[score_key])
    
    return df


from pandas import read_csv
from numpy import unique
import pathlib
from pandas import DataFrame


def select_resource(resource_name: str = 'consensus') -> DataFrame:
    """
    Read resource of choice from the pre-generated resources in LIANA.

    Parameters
    ----------
    resource_name
        Name of the resource to be loaded and use for ligand-receptor inference.

    Returns
    -------
    A dataframe with ``['ligand', 'receptor']`` columns

    """

    resource_name = resource_name.lower()

    resource_path = pathlib.Path(__file__).parent.joinpath("omni_resource.csv")
    resource = read_csv(resource_path, index_col=False)

    resource = resource[resource['resource'] == resource_name]

    resource = resource[['source_genesymbol', 'target_genesymbol']]
    resource = resource.rename(columns={'source_genesymbol': 'ligand',
                                        'target_genesymbol': 'receptor'})

    return resource


def show_resources():
    """
    Show provided resources.

    Returns
    -------
    A list of resource names available via ``liana.resource.select_resource``

    """
    resource_path = pathlib.Path(__file__).parent.joinpath("omni_resource.csv")
    resource = read_csv(resource_path, index_col=False)
    return list(unique(resource.resource))



from liana.method._Method import MethodMeta
from liana.method._liana_pipe import liana_pipe

from anndata import AnnData
from pandas import DataFrame, concat
from typing import Optional


class AggregateClass(MethodMeta):
    """LIANA's Method Consensus Class"""
    def __init__(self, _SCORE, methods):
        super().__init__(method_name=_SCORE.method_name,
                         complex_cols=[],
                         add_cols=[],
                         fun=_SCORE.fun,
                         magnitude=_SCORE.magnitude,
                         magnitude_ascending=True,
                         specificity=_SCORE.specificity,
                         specificity_ascending=True,
                         permute=_SCORE.permute,
                         reference=_SCORE.reference
                         )
        self._SCORE = _SCORE
        self.methods = methods
        self.steady = 'steady_rank'
        self.steady_ascending = True

        # Define sc to aggregate
        self.specificity_specs = {method.method_name: (
            method.specificity, method.specificity_ascending) for method in methods
            if method.specificity is not None}
        self.magnitude_specs = {method.method_name: (
            method.magnitude, method.magnitude_ascending) for method in methods
            if method.magnitude is not None}

        # If SingleCellSignalR is in there also add it to calculate method; TODO just remove this
        methods_by_name = {method.method_name: method for method in methods}
        if 'SingleCellSignalR' in methods_by_name.keys():
            self.steady_specs = self.specificity_specs.copy()
            self.steady_specs['SingleCellSignalR'] = \
                (methods_by_name['SingleCellSignalR'].magnitude,
                 methods_by_name['SingleCellSignalR'].magnitude_ascending)

        # Define additional columns needed depending on the methods to be run
        self.add_cols = list(
            {x for li in [method.add_cols for method in methods] for x in li}
        )
        self.complex_cols = list(
            {x for li in [method.complex_cols for method in methods] for x in li}
        )

    def describe(self):
        """Briefly described the method"""
        print(
            f"{self.method_name} returns `{self.magnitude}`, `{self.specificity}`,"
            f" and `{self.steady}`. "
            f"{self.magnitude} and {self.specificity} represent an aggregate of the "
            f"`magnitude`- and `specificity`-related scoring functions from the different methods."
            f"{self.steady} (DEPRECATED) represents one scoring function from each method intended"
            f" to prioritize ligand-receptor interactions in steady-state data, "
            f"regardless if they represent `specificity` or `magnitude`."
        )

    def __call__(self,
                 adata: AnnData,
                 groupby: str,
                 resource_name: str = 'consensus',
                 expr_prop: float = 0.1,
                 min_cells: int = 5,
                 base: float = 2.718281828459045,
                 aggregate_method='rra',
                 return_all_lrs: bool = False,
                 key_added : str = 'liana_res',
                 consensus_opts=None,
                 use_raw: Optional[bool] = True,
                 layer: Optional[str] = None,
                 de_method='t-test',
                 verbose: Optional[bool] = False,
                 n_perms: int = 1000,
                 seed: int = 1337,
                 resource: Optional[DataFrame] = None,
                 inplace=True
                 ):
        """
        Parameters
        ----------
        adata
            Annotated data object.
        groupby
            The key of the observations grouping to consider.
        resource_name
            Name of the resource to be loaded and use for ligand-receptor inference.
        expr_prop
            Minimum expression proportion for the ligands/receptors (and their subunits) in the
             corresponding cell identities. Set to `0`, to return unfiltered results.
        min_cells
            Minimum cells per cell identity (`groupby`) to be considered for downstream analysis
        base
            Exponent base used to reverse the log-transformation of matrix. Note that this is
            relevant only for the `logfc` method.
        aggregate_method
            Method aggregation approach, one of ['mean', 'rra'], where `mean` represents the
            mean rank, while 'rra' is the RobustRankAggregate (Kolde et al., 2014)
            of the interactions
        return_all_lrs
            Bool whether to return all LRs, or only those that surpass the `expr_prop`
            threshold. Those interactions that do not pass the `expr_prop` threshold will
            be assigned to the *worst* score of the ones that do. `False` by default.
        key_added
            Key to add the results to the `uns` attribute of `adata`.
        use_raw
            Use raw attribute of adata if present. True, by default.
        layer
            Layer in anndata.AnnData.layers to use. If None, use anndata.AnnData.X.
        de_method
            Differential expression method. `scanpy.tl.rank_genes_groups` is used to rank genes
            according to 1vsRest. The default method is 't-test'.
        verbose
            Verbosity flag
        n_perms
            Number of permutations for the permutation test. Note that this is relevant
            only for permutation-based methods - e.g. `CellPhoneDB`
        seed
            Random seed for reproducibility.
        resource
            Parameter to enable external resources to be passed. Expects a pandas dataframe
            with [`ligand`, `receptor`] columns. None by default. If provided will overrule
            the resource requested via `resource_name`
        inplace
            If true return `DataFrame` with results, else assign inplace to `.uns`.


        Returns
        -------
        If ``inplace = False``, returns a `DataFrame` with ligand-receptor results
        Otherwise, modifies the ``adata`` object with the following key:
            - :attr:`anndata.AnnData.uns` ``['liana_res']`` with the aforementioned DataFrame
        """
        liana_res = liana_pipe(adata=adata,
                               groupby=groupby,
                               resource_name=resource_name,
                               resource=resource,
                               expr_prop=expr_prop,
                               min_cells=min_cells,
                               base=base,
                               return_all_lrs=return_all_lrs,
                               de_method=de_method,
                               verbose=verbose,
                               _score=self,
                               use_raw=use_raw,
                               layer=layer,
                               n_perms=n_perms,
                               seed=seed,
                               _methods=self.methods,
                               _aggregate_method=aggregate_method,
                               _consensus_opts=consensus_opts
                               )
        adata.uns[key_added] = liana_res

        return None if inplace else liana_res

_rank_aggregate_meta = \
    MethodMeta(method_name="Rank_Aggregate",
               complex_cols=[],
               add_cols=[],
               fun=None,  # change to _robust_rank
               magnitude='magnitude_rank',
               magnitude_ascending=True,
               specificity='specificity_rank',
               specificity_ascending=True,
               permute=False,
               reference='Dimitrov, D., Türei, D., Garrido-Rodriguez, M., Burmedi, P.L., '
                         'Nagai, J.S., Boys, C., Ramirez Flores, R.O., Kim, H., Szalai, B., '
                         'Costa, I.G. and Valdeolivas, A., 2022. Comparison of methods and '
                         'resources for cell-cell communication inference from single-cell '
                         'RNA-Seq data. Nature Communications, 13(1), pp.1-13. '
               )



import numpy as np

from liana.method._Method import Method, MethodMeta


def _sca_score(x):
    """
    Calculate SingleCellSignalR-like LRscore

    Parameters
    ----------
    x
        DataFrame row

    Returns
    -------
    (LRscore, None)

    """

    lr_sqrt = np.sqrt(x.ligand_means) * np.sqrt(x.receptor_means)
    return lr_sqrt / (lr_sqrt + x.mat_mean), None


# Initialize CPDB Meta
_singlecellsignalr = MethodMeta(method_name="SingleCellSignalR",
                                complex_cols=['ligand_means', 'receptor_means'],
                                add_cols=['mat_mean'],
                                fun=_sca_score,
                                magnitude='lrscore',
                                magnitude_ascending=False,
                                specificity=None,
                                specificity_ascending=None,
                                permute=False,
                                reference='Cabello-Aguilar, S., Alame, M., Kon-Sun-Tack, F., Fau, '
                                          'C., Lacroix, M. and Colinge, J., '
                                          '2020. SingleCellSignalR: inference of intercellular '
                                          'networks from single-cell transcriptomics. Nucleic '
                                          'Acids Research, 48(10), pp.e55-e55. '
                                )

# Initialize callable Method instance
singlecellsignalr = Method(_SCORE=_singlecellsignalr)



import numpy as np
import pandas as pd
from functools import reduce
from scipy.stats import rankdata, beta


def _aggregate(lrs: dict,
               consensus,
               aggregate_method: str = 'rra',
               consensus_opts: list = None,
               _key_cols: list = None
               ) -> pd.DataFrame:
    """

    Parameters
    ---------
    lrs
        a list with results for all methods
    consensus
        ConsensusClass instance used to generate the lr results
    _key_cols
        should represent unique LRs columns by which to join
    aggregate_method
        method by which we aggregate the ranks. Options are ['rra', 'mean'],
        where 'rra' corresponds to the RRA method, 'mean' is just the mean of the ranks.
    consensus_opts
        consensus ranks to be obtained

    Returns
    -------
    A long pd.DataFrame with ranked LRs
    """

    # join the sc to the whole universe between the methods
    if _key_cols is None:
        _key_cols = ['source', 'target', 'ligand_complex', 'receptor_complex']
    if consensus_opts is None:
        consensus_opts = ['Steady', 'Magnitude', 'Specificity']

    lrs = [lrs[method].drop_duplicates(keep='first') for method in lrs]
    # reduce to a df with the shared keys + all relevant sc
    lr_res = reduce(
        lambda left, right:
        pd.merge(left, right, how='outer', on=_key_cols,
                 suffixes=('', '_duplicated')), lrs
    )
    # drop duplicated columns
    lr_res = lr_res.loc[:, ~lr_res.columns.str.endswith('_duplicated')]

    order_col = ''
    if 'Steady' in consensus_opts:
        lr_res[consensus.steady] = _rank_aggregate(lr_res.copy(),
                                                   consensus.steady_specs,
                                                   _key_cols,
                                                   aggregate_method=aggregate_method)
        order_col = consensus.steady
    if 'Specificity' in consensus_opts:
        lr_res[consensus.specificity] = _rank_aggregate(lr_res.copy(),
                                                        consensus.specificity_specs,
                                                        _key_cols,
                                                        aggregate_method=aggregate_method)
        order_col = consensus.specificity
    if 'Magnitude' in consensus_opts:
        lr_res[consensus.magnitude] = _rank_aggregate(lr_res.copy(),
                                                      consensus.magnitude_specs,
                                                      _key_cols,
                                                      aggregate_method=aggregate_method)
        order_col = consensus.magnitude

    lr_res = lr_res.sort_values(order_col)

    return lr_res


def _rank_aggregate(lr_res, specs, _key_cols, aggregate_method) -> np.array:
    """
    Aggregate method ranks

    Parameters
    ----------
    lr_res
        joined results from all methods
    specs
        specs dictionary where method_name:(score_name, score_desc)
    _key_cols
        columns by which we join
    aggregate_method
        method by which to aggregate the ranks

    Returns
    -------
    An array of values /w length of lr_res.shape[0]

    """
    assert aggregate_method in ['rra', 'mean']

    # Convert specs columns to ranks
    for spec in specs:
        score_name = specs[spec][0]
        ascending = specs[spec][1]

        if ascending:
            lr_res[score_name] = rankdata(lr_res[score_name], method='average')
        else:
            lr_res[score_name] = rankdata(lr_res[score_name] * -1,
                                          method='average')

    # get only the relevant ranks as a mat (joins order the keys)
    scores = list({specs[s][0] for s in specs})
    rmat = lr_res[scores].values

    if aggregate_method == 'rra':
        return _robust_rank_aggregate(rmat)
    elif aggregate_method == 'mean':
        return np.mean(rmat, axis=1)


def _corr_beta_pvals(p, k) -> np.array:
    """
    Correct beta p-values

    Parameters
    ----------
    p
        (min) p-value
    k
        total number of rows

    Returns
    -------
    An array with corrected p-values
    """
    p = np.clip(p * k, a_min=0, a_max=1)
    return p


def _rho_scores(rmat, dist_a, dist_b):
    """
    Calculate Beta Distribution Rho Scores
    
    ----------
    rmat
        a matrix where rows are the ranks/n for each interaction, while
        columns correspond to each method
    dist_a
        non-negative shape param a
    dist_b
        non-negative shape param b

    Returns
    -------
    A vector of pvals as implemented in the RRA method
    """

    # Sort values by sources (rows)
    rmat = np.sort(rmat, axis=1)
    # Calc beta cdf across rows
    p = beta.cdf(rmat, dist_a, dist_b)
    # get min pval per row
    p = np.min(p, axis=1)
    # correct p-vals
    rho = _corr_beta_pvals(p, k=rmat.shape[1])

    return rho


def _robust_rank_aggregate(rmat) -> np.array:
    """
    Calculate Robust Rank Aggregate as in Kolde et al., 2012

    Parameters
    ----------
    rmat
        Matrix with interaction ranks (rows) for each method (columns)

    Returns
    -------
    An array with p-values for each row
    """

    # 0-1 values depending on relative rank of
    # each interaction divided by the max of each method
    # due to max diffs due to ties
    rmat = rmat / np.max(rmat, axis=0)
    # generate dist_a/b with same row size as rmat
    dist_a = np.repeat([np.arange(rmat.shape[1])], rmat.shape[0], axis=0) + 1
    dist_b = rmat.shape[1] - dist_a + 1

    return _rho_scores(rmat, dist_a, dist_b)



from liana.resource import select_resource
from liana.funcomics import generate_lr_geneset

import decoupler as dc


def test_generate_lr_resource():
    """Test generate_lr_resource."""
    # load data
    net = dc.get_progeny(top=1000, organism='human') # reduce top to 1000 for testing
    resource = select_resource('consensus')
    
    lr_net = generate_lr_geneset(resource, net)
    
    # check if the result has the right columns
    assert set(lr_net.columns) == set(['interaction', 'weight', 'source'])
    
    # check if the result has the right number of rows
    assert lr_net.shape[0] == 177
    
    # check if the result has the right number of unique interactions
    assert lr_net['interaction'].nunique() == 158
    
    # check if the result has the right number of unique sources
    assert lr_net['source'].nunique() == 14
    
    # check the weight of a specific interaction
    assert lr_net[lr_net['interaction'] == 'LAMB3&ITGAV_ITGB8']['weight'].values[0] == 3.6229854822158813


# Configuration file for the Sphinx documentation builder.

import os
import sys

sys.path.insert(0, os.path.abspath('../..'))


def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


project = 'liana-py'
copyright = '2022, Saezlab'
author = 'Daniel Dimitrov'
release = get_version("../../liana/__init__.py")
version = get_version("../../liana/__init__.py")

# -- General configuration
extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosectionlabel',
    'numpydoc',
    'nbsphinx',
    'IPython.sphinxext.ipython_console_highlighting'
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output
master_doc = 'index'

html_theme = 'sphinx_rtd_theme'
html_static_path = ["_static"]
html_theme_options = dict(
    logo_only=True,
    display_version=True,
)
html_context = dict(
    display_github=False,
    github_user='saezlab',
    github_repo='liana-py',
    github_version='main',
    conf_py_path='/docs/source/',
)
html_show_sphinx = True
html_logo = 'logo.png'
html_favicon = 'logo.png'
html_css_files = ['custom.css']

# -- Options for EPUB output
epub_show_urls = 'footnote'

nbsphinx_execute = 'never'



from __future__ import annotations

import anndata
import pandas

from liana.method._pipe_utils import prep_check_adata, assert_covered, filter_resource, \
    filter_reassemble_complexes
from ..resource import select_resource, explode_complexes
from liana.method._pipe_utils._get_mean_perms import _get_means_perms
from liana.method._pipe_utils._aggregate import _aggregate

import scanpy as sc
import pandas as pd
import numpy as np
from functools import reduce


def liana_pipe(adata: anndata.AnnData,
               groupby: str,
               resource_name: str,
               resource: pd.DataFrame | None,
               expr_prop: float,
               min_cells: int,
               base: float,
               de_method: str,
               n_perms: int,
               seed: int,
               verbose: bool,
               use_raw: bool,
               layer: str | None,
               supp_columns: list | None = None,
               return_all_lrs: bool = False,
               _key_cols: list = None,
               _score=None,
               _methods: list = None,
               _consensus_opts: list = None,
               _aggregate_method: str = None
               ):
    """
    Parameters
    ----------
    adata
        Annotated data object.
    groupby
        The key of the observations grouping to consider.
    resource_name
        Name of the resource to be loaded and use for ligand-receptor inference.
    resource
        Parameter to enable external resources to be passed. Expects a pandas dataframe
        with [`ligand`, `receptor`] columns. None by default. If provided will overrule
        the resource requested via `resource_name`
    expr_prop
        Minimum expression proportion for the ligands/receptors (and their subunits) in the
         corresponding cell identities. Set to `0`, to return unfiltered results.
    min_cells
        Minimum cells per cell identity
    base
        The base by which to do expm1 (relevant only for 1vsRest logFC calculation)
    de_method
        Differential expression method. `scanpy.tl.rank_genes_groups` is used to rank genes
        according to 1vsRest. The default method is 't-test'. Only relevant if p-values
        are included in `supp_cols`
    n_perms
        n permutations (relevant only for permutation-based methods)
    seed
        Random seed for reproducibility 
    verbose
        Verbosity flag
    use_raw
        Use raw attribute of adata if present.
    layer
        Layer in anndata.AnnData.layers to use. If None, use anndata.AnnData.X.
    supp_columns
        additional columns to be added to the output of each method.
    return_all_lrs
        Bool whether to return all LRs, or only those that surpass the expr_prop threshold.
        `False` by default.
    _key_cols
        columns which make every interaction unique (i.e. PK).
    _score
        Instance of Method classes (None by default - returns LR stats - no methods used).
    _methods
        Methods to be run (only relevant for consensus).
    _consensus_opts
        Ways to aggregate interactions across methods by default does all aggregations (['Steady',
        'Specificity', 'Magnitude']).
    _aggregate_method
        RobustRankAggregate('rra') or mean rank ('mean').

    Returns
    -------
    A adata frame with ligand-receptor results

    """
    if _key_cols is None:
        _key_cols = ['source', 'target', 'ligand_complex', 'receptor_complex']

    if _score is not None:
        _complex_cols, _add_cols = _score.complex_cols, _score.add_cols
    else:
        _complex_cols = ['ligand_means', 'receptor_means']
        # change to full list and move to _var
        _add_cols = ['ligand_means_sums', 'receptor_means_sums',
                     'ligand_zscores', 'receptor_zscores',
                     'ligand_logfc', 'receptor_logfc',
                     'ligand_trimean', 'receptor_trimean',
                     'mat_mean', 'mat_max',
                     ]

    if supp_columns is None:
        supp_columns = []
    _add_cols = _add_cols + ['ligand', 'receptor', 'ligand_props', 'receptor_props'] + supp_columns

    # initialize mat_mean for sca
    mat_mean = None
    mat_max = None

    # Check and Reformat Mat if needed
    adata = prep_check_adata(adata=adata,
                             groupby=groupby,
                             min_cells=min_cells,
                             use_raw=use_raw,
                             layer=layer,
                             verbose=verbose)

    # get mat mean for SCA (before reducing the features)
    if 'mat_mean' in _add_cols:
        mat_mean = np.mean(adata.X, dtype='float32')

    # get mat max for CellChat
    if 'mat_max' in _add_cols:
        mat_max = np.max(adata.X.data)
        assert isinstance(mat_max, np.float32)

    if resource is None:
        resource = select_resource(resource_name.lower())
    # explode complexes/decomplexify
    resource = explode_complexes(resource)

    # Check overlap between resource and adata
    assert_covered(np.union1d(np.unique(resource["ligand"]),
                              np.unique(resource["receptor"])),
                   adata.var_names, verbose=verbose)

    # Filter Resource
    resource = filter_resource(resource, adata.var_names)

    if verbose:
        print(f"Generating ligand-receptor stats for {adata.shape[0]} samples "
              f"and {adata.shape[1]} features")

    # Create Entities
    entities = np.union1d(np.unique(resource["ligand"]),
                          np.unique(resource["receptor"]))

    # Filter to only include the relevant genes
    adata = adata[:, np.intersect1d(entities, adata.var.index)]

    # Get lr results
    lr_res = _get_lr(adata=adata, resource=resource,
                     mat_mean=mat_mean, mat_max=mat_max,
                     relevant_cols=_key_cols + _add_cols + _complex_cols,
                     de_method=de_method, base=base, verbose=verbose)

    # Mean Sums required for NATMI (note done on subunits also)
    if 'ligand_means_sums' in _add_cols:
        lr_res = _sum_means(lr_res, what='ligand_means',
                            on=['ligand_complex', 'receptor_complex',
                                'ligand', 'receptor', 'target'])
    if 'receptor_means_sums' in _add_cols:
        lr_res = _sum_means(lr_res, what='receptor_means',
                            on=['ligand_complex', 'receptor_complex',
                                'ligand', 'receptor', 'source'])

    # Calculate Score
    if _score is not None:
        if _score.method_name == "Rank_Aggregate":
            # Run all methods in consensus
            lrs = {}
            for method in _methods:
                if verbose:
                    print(f"Running {method.method_name}")

                lrs[method.method_name] = \
                    _run_method(lr_res=lr_res.copy(),
                                adata=adata,
                                expr_prop=expr_prop,
                                _score=method,
                                _key_cols=_key_cols,
                                _complex_cols=method.complex_cols,
                                _add_cols=method.add_cols,
                                n_perms=n_perms,
                                seed=seed,
                                return_all_lrs=return_all_lrs,
                                verbose=verbose,
                                _aggregate_flag=True
                                )
            if _consensus_opts is not False:
                lr_res = _aggregate(lrs,
                                    consensus=_score,
                                    aggregate_method=_aggregate_method,
                                    _key_cols=_key_cols)
            else:  # Return by method results as they are
                return lrs
        else:  # Run the specific method in mind
            lr_res = _run_method(lr_res=lr_res, adata=adata, expr_prop=expr_prop,
                                 _score=_score, _key_cols=_key_cols, _complex_cols=_complex_cols,
                                 _add_cols=_add_cols, n_perms=n_perms,
                                 return_all_lrs=return_all_lrs,
                                 verbose=verbose, seed=seed)
    else:  # Just return lr_res
        lr_res = filter_reassemble_complexes(lr_res=lr_res,
                                             _key_cols=_key_cols,
                                             expr_prop=expr_prop,
                                             complex_cols=_complex_cols,
                                             return_all_lrs=return_all_lrs)

    return lr_res


def _join_stats(source, target, dedict, resource):
    """
    Joins and renames source-ligand and target-receptor stats to the ligand-receptor resource

    Parameters
    ----------
    source
        Source/Sender cell type
    target
        Target/Receiver cell type
    dedict
        dictionary
    resource
        Ligand-receptor Resource

    Returns
    -------
    Ligand-Receptor stats

    """
    source_stats = dedict[source].copy()
    source_stats.columns = source_stats.columns.map(
        lambda x: 'ligand_' + str(x))
    source_stats = source_stats.rename(
        columns={'ligand_names': 'ligand', 'ligand_label': 'source'})

    target_stats = dedict[target].copy()
    target_stats.columns = target_stats.columns.map(
        lambda x: 'receptor_' + str(x))
    target_stats = target_stats.rename(
        columns={'receptor_names': 'receptor', 'receptor_label': 'target'})

    bound = resource.merge(source_stats).merge(target_stats)

    return bound


def _get_lr(adata, resource, relevant_cols, mat_mean, mat_max, de_method, base, verbose):
    """
    Run DE analysis and merge needed information with resource for LR inference

    Parameters
    ----------
    adata : anndata.AnnData
        adata filtered and formated to contain only the relevant features for lr inference

    relevant_cols : list
        Relevant column names

    resource : pandas.core.frame.DataFrame formatted and filtered resource
    dataframe with the following columns: [interaction, ligand, receptor,
    ligand_complex, receptor_complex]

    de_method : Literal name of the DE `method` to call via
    scanpy.tl.rank_genes.groups - available options are: ['logreg', 't-test',
    'wilcoxon', 't-test_overestim_var']

    Returns
    -------
    lr_res : pandas.core.frame.DataFrame long-format pandas dataframe with stats
    for all interactions /w matching variables in the dataset.

    """
    # get label cats
    labels = adata.obs.label.cat.categories

    # Method-specific stats
    connectome_flag = ('ligand_zscores' in relevant_cols) | (
            'receptor_zscores' in relevant_cols)
    if connectome_flag:
        adata.layers['scaled'] = sc.pp.scale(adata, copy=True).X

    logfc_flag = ('ligand_logfc' in relevant_cols) | (
            'receptor_logfc' in relevant_cols)
    if logfc_flag:
        if 'log1p' in adata.uns_keys():
            if (adata.uns['log1p']['base'] is not None) & verbose:
                print("Assuming that counts were `natural` log-normalized!")
        elif ('log1p' not in adata.uns_keys()) & verbose:
            print("Assuming that counts were `natural` log-normalized!")
        adata.layers['normcounts'] = adata.X.copy()
        adata.layers['normcounts'].data = _expm1_base(adata.X.data, base)

    # initialize dict
    dedict = {}

    # Calc pvals + other stats per gene or not
    rank_genes_bool = ('ligand_pvals' in relevant_cols) | ('receptor_pvals' in relevant_cols)
    if rank_genes_bool:
        adata = sc.tl.rank_genes_groups(adata, groupby='label',
                                        method=de_method, use_raw=False,
                                        copy=True)

    for label in labels:
        temp = adata[adata.obs.label == label, :]
        a = _get_props(temp.X)
        stats = pd.DataFrame({'names': temp.var_names, 'props': a}). \
            assign(label=label).sort_values('names')
        if rank_genes_bool:
            pvals = sc.get.rank_genes_groups_df(adata, label)
            stats = stats.merge(pvals)
        dedict[label] = stats

    # check if genes are ordered correctly
    if not list(adata.var_names) == list(dedict[labels[0]]['names']):
        raise AssertionError("Variable names did not match DE results!")

    # Calculate Mean, logFC and z-scores by group
    for label in labels:
        temp = adata[adata.obs.label.isin([label])]
        dedict[label]['means'] = temp.X.mean(axis=0).A.flatten()
        if connectome_flag:
            dedict[label]['zscores'] = temp.layers['scaled'].mean(axis=0)
        if logfc_flag:
            dedict[label]['logfc'] = _calc_log2fc(adata, label)
        if isinstance(mat_max, np.float32):  # cellchat flag
            dedict[label]['trimean'] = _trimean(temp.X / mat_max)

    # Create df /w cell identity pairs
    pairs = (pd.DataFrame(np.array(np.meshgrid(labels, labels))
                          .reshape(2, np.size(labels) * np.size(labels)).T)
             .rename(columns={0: "source", 1: "target"}))

    # Join Stats
    lr_res = pd.concat(
        [_join_stats(source, target, dedict, resource) for source, target in
         zip(pairs['source'], pairs['target'])]
    )

    if 'mat_mean' in relevant_cols:
        assert isinstance(mat_mean, np.float32)
        lr_res['mat_mean'] = mat_mean

    if isinstance(mat_max, np.float32):
        lr_res['mat_max'] = mat_max

    # subset to only relevant columns
    relevant_cols = np.intersect1d(relevant_cols, lr_res.columns)

    return lr_res[relevant_cols]


def _sum_means(lr_res, what, on):
    """
    Calculate Sum Means

    Parameters
    ---------
    lr_res
        lr_res with reassembled complexes
    what
        [entity]_means_sums for which the sum is calculated
    on
        columns by which to group and sum

    Returns
    -------
    lr_res with [entity]_means_sums column
    """
    return lr_res.join(lr_res.groupby(on)[what].sum(), on=on, rsuffix='_sums')


def _calc_log2fc(adata, label) -> np.ndarray:
    """
    Calculate 1 vs rest Log2FC for a particular cell identity

    Parameters
    ---------

    adata
        anndata with feature-space reduced to the vars intersecting with the IDs in the resource
    label
        cell identity

    Returns
    -------
    An array with logFC changes

    """
    # Get subject vs rest cells
    subject = adata[adata.obs.label.isin([label])]
    rest = adata[~adata.obs.label.isin([label])]

    # subject and rest means
    subj_means = subject.layers['normcounts'].mean(0).A.flatten()
    rest_means = rest.layers['normcounts'].mean(0).A.flatten()

    # log2 + 1 transform
    subj_log2means = np.log2(subj_means + 1)
    loso_log2means = np.log2(rest_means + 1)

    logfc_vec = subj_log2means - loso_log2means

    return logfc_vec


# Exponent with a custom base
def _expm1_base(X, base):
    return np.power(base, X) - 1


def _run_method(lr_res: pandas.DataFrame,
                adata: anndata.AnnData,
                expr_prop: float,
                _score,
                _key_cols: list,
                _complex_cols: list,
                _add_cols: list,
                n_perms: int,
                seed: int,
                return_all_lrs: bool,
                verbose: bool,
                _aggregate_flag: bool = False  # Indicates whether we're generating the consensus
                ) -> pd.DataFrame:
    # re-assemble complexes - specific for each method
    lr_res = filter_reassemble_complexes(lr_res=lr_res,
                                         _key_cols=_key_cols,
                                         expr_prop=expr_prop,
                                         return_all_lrs=return_all_lrs,
                                         complex_cols=_complex_cols)

    _add_cols = _add_cols + ['ligand', 'receptor']
    relevant_cols = reduce(np.union1d, [_key_cols, _complex_cols, _add_cols])
    if return_all_lrs:
        relevant_cols = list(relevant_cols) + ['lrs_to_keep']
        # separate those that pass from rest
        rest_res = lr_res[~lr_res['lrs_to_keep']]
        rest_res = rest_res[relevant_cols]
        lr_res = lr_res[lr_res['lrs_to_keep']]
    lr_res = lr_res[relevant_cols]

    if ('mat_max' in _add_cols) & (_score.method_name == "CellChat"):
        # CellChat matrix_max
        norm_factor = np.unique(lr_res['mat_max'].values).item()
        agg_fun = _trimean
    else:
        norm_factor = None
        agg_fun = np.mean

    if _score.permute:
        perms, ligand_pos, receptor_pos, labels_pos = \
            _get_means_perms(adata=adata,
                             lr_res=lr_res,
                             n_perms=n_perms,
                             seed=seed,
                             agg_fun=agg_fun,
                             norm_factor=norm_factor,
                             verbose=verbose)
        lr_res[[_score.magnitude, _score.specificity]] = \
            lr_res.apply(_score.fun, axis=1, result_type="expand",
                         perms=perms, ligand_pos=ligand_pos,
                         receptor_pos=receptor_pos, labels_pos=labels_pos)
    else:  # non-perm funs
        lr_res[[_score.magnitude, _score.specificity]] = \
            lr_res.apply(_score.fun, axis=1, result_type="expand")

    if return_all_lrs:
        # re-append rest of results
        lr_res = pd.concat([lr_res, rest_res], copy=False)
        if _score.magnitude is not None:
            fill_value = _assign_min_or_max(lr_res[_score.magnitude],
                                            _score.magnitude_ascending)
            lr_res.loc[~lr_res['lrs_to_keep'], _score.magnitude] = fill_value
        if _score.specificity is not None:
            fill_value = _assign_min_or_max(lr_res[_score.specificity],
                                            _score.specificity_ascending)
            lr_res.loc[~lr_res['lrs_to_keep'], _score.specificity] = fill_value

    if _aggregate_flag:  # if consensus keep only the keys and the method scores
        lr_res = lr_res[_key_cols + [_score.magnitude, _score.specificity]]

    # remove redundant cols for some scores
    if (_score.magnitude is None) | (_score.specificity is None):
        lr_res = lr_res.drop([None], axis=1)

    return lr_res


def _assign_min_or_max(x, x_ascending):
    if x_ascending:
        return np.max(x)
    else:
        return np.min(x)


# Function to get gene expr proportions
def _get_props(X_mask):
    return X_mask.getnnz(axis=0) / X_mask.shape[0]


def _trimean(a, axis=0):
    """
    Tukey's mean / Trimean
    Parameters
    ----------
    a
        array (cell ident counts)

    Returns
    -------
    Trimean value

    """
    return np.mean(
        np.quantile(a.A,
                    q=[0.25, 0.5, 0.5, 0.75],
                    axis=axis),
        axis=axis)



"""
Functions to deal with protein complexes
"""
from __future__ import annotations

import pandas as pd
import numpy as np


def filter_reassemble_complexes(lr_res,
                                _key_cols,
                                complex_cols,
                                expr_prop,
                                return_all_lrs=False,
                                complex_policy='min'):
    """
    Reassemble complexes from exploded long-format pandas Dataframe.

    Parameters
    ----------
    lr_res
        long-format pandas dataframe with exploded complex subunits
    _key_cols
        primary key for lr_res, typically a list with the following elements -
        ['source', 'target', 'ligand_complex', 'receptor_complex']
    complex_cols
        method/complex-relevant columns
    expr_prop
        minimum expression proportion for each subunit in a complex
    return_all_lrs
        Bool whether to return all LRs, or only those that surpass the expr_prop
        threshold. `False` by default.
    complex_policy
        approach by which the complexes are reassembled

    Return
    -----------
    :return: lr_res: a reduced long-format pandas dataframe
    """
    # Filter by expr_prop (inner join only complexes where all subunits are expressed)
    expressed = (lr_res[_key_cols + ['ligand_props', 'receptor_props']]
                 .set_index(_key_cols)
                 .stack()
                 .groupby(_key_cols)
                 .agg(prop_min=complex_policy)
                 .reset_index()
                 )
    expressed = expressed[expressed['prop_min'] >= expr_prop]

    if not return_all_lrs:
        lr_res = lr_res.merge(expressed, how='inner', on=_key_cols)
    else:
        expressed['lrs_to_keep'] = True
        lr_res = lr_res.merge(expressed, how='left', on=_key_cols)
         # deal with duplicated subunits
         # subunits that are not expressed might not represent the most relevant subunit
        lr_res.drop_duplicates(subset=_key_cols, inplace=True)
        lr_res['lrs_to_keep'].fillna(value=False, inplace=True)
        lr_res['prop_min'].fillna(value=0, inplace=True)

    # check if complex policy is only min
    aggs = {complex_policy, 'min'}
    
    for col in complex_cols:
        lr_res = _reduce_complexes(col=col, 
                                   lr_res=lr_res,
                                   key_cols=_key_cols,
                                   aggs=aggs)


    # check if there are any duplicated subunits
    duplicate_mask = lr_res.duplicated(subset=_key_cols, keep=False)
    if duplicate_mask.any():
        # check if there are any non-equal subunit values
        if not lr_res[duplicate_mask].groupby(_key_cols)[complex_cols].transform(lambda x: x.duplicated(keep=False)).all().all():
            print('Warning: there were duplicated subunits in the complexes. ' + 
                  'The subunits were reduced to only the minimum expression subunit. ' +
                  'However, there were subunits that were not the same within a complex. ')
        lr_res = lr_res.drop_duplicates(subset=_key_cols, keep='first')

    return lr_res



def _reduce_complexes(col: str,
                      lr_res: pd.DataFrame,
                      key_cols: list,
                      aggs: (dict | str)
                      ):
    """
    Reduce the complexes

    Parameters
    ------------

    col
        column by which we are reducing the complexes
    lr_res
     liana_pipe generated long DataFrame
    key_cols
        a list of columns that define each row as unique
    aggs
        dictionary with the way(s) by which we aggregate. Note 'min' should be there -
        we need the miniumum to find the lowest expression subunits, which are then used
        to reduce the exploded complexes

    Return
    -----------
    :returns:
        lr_res with exploded complexes reduced to only the minimum (default) subunit

    """    
    # Group by keys
    lr_res = lr_res.groupby(key_cols)

    # Get min cols by which we will join
    # then rename from agg name to column name (e.g. 'min' to 'ligand_min')
    lr_min = lr_res[col].agg(aggs).reset_index().copy(). \
        rename(columns={agg: col.split('_')[0] + '_' + agg for agg in aggs})

    # right is the min subunit for that column
    join_key = col.split('_')[0] + '_min'  # ligand_min or receptor_min

    # Here, I join the min value and keep only those rows that match
    lr_res = lr_res.obj.merge(lr_min, on=key_cols, how='inner')
    lr_res = lr_res[lr_res[col] == lr_res[join_key]].drop(join_key, axis=1)

    return lr_res



from __future__ import annotations

import anndata
import numpy as np
import pandas

from plotnine import ggplot, geom_point, aes, \
    facet_grid, labs, theme_bw, theme, element_text, element_rect, scale_size_continuous


def dotplot(adata: anndata.AnnData = None,
            uns_key = 'liana_res',
            liana_res: pandas.DataFrame = None,
            colour: str = None,
            size: str = None,
            source_labels: list = None,
            target_labels: list = None,
            top_n: int = None,
            orderby: str | None = None,
            orderby_ascending: bool | None = None,
            filterby: bool | None = None,
            filter_lambda=None,
            inverse_colour: bool = False,
            inverse_size: bool = False,
            size_range: tuple = (2, 9),
            figure_size: tuple = (8, 6),
            return_fig=True) -> ggplot:
    """
    Dotplot interactions by source and target cells

    Parameters
    ----------
    adata
        `AnnData` object with `uns_key` in `adata.uns`. Default is `None`.
    uns_key
        key in adata.uns where liana_res is stored. Defaults to 'liana_res'.
    liana_res
        `liana_res` a `DataFrame` in liana's format
    colour
        `column` in `liana_res` to define the colours of the dots
    size
        `column` in `liana_res` to define the size of the dots
    source_labels
        list to specify `source` identities to plot
    target_labels
        list to specify `target` identities to plot
    top_n
        Obtain only the top_n interactions to plot. Default is `None`
    orderby
        If `top_n` is not `None`, order the interactions by these columns
    orderby_ascending
        If `top_n` is not `None`, specify how to order the interactions
    filterby
        Column by which to filter the interactions
    filter_lambda
        If `filterby` is not `None`, provide a simple lambda function by which
        to filter the interactions to be plotted
    inverse_colour
        Whether to -log10 the `colour` column for plotting. `False` by default.
    inverse_size
        Whether to -log10 the `size` column for plotting. `False` by default.
    size_range
        Define size range - (min, max). Default is (2, 9)
    figure_size
        Figure x,y size
    return_fig
        `bool` whether to return the fig object, `False` only plots

    Returns
    -------
    A `plotnine.ggplot` instance

    """
    liana_res = _prep_liana_res(adata=adata,
                                liana_res=liana_res, 
                                source_labels=source_labels,
                                target_labels=target_labels,
                                size=size,
                                colour=colour,
                                uns_key=uns_key)

    if filterby is not None:
        msk = liana_res[filterby].apply(filter_lambda)
        relevant_interactions = np.unique(liana_res[msk].interaction)
        liana_res = liana_res[np.isin(liana_res.interaction, relevant_interactions)]

    if top_n is not None:
        # get the top_n for each interaction
        if orderby is None:
            ValueError("Please specify the column to order the interactions.")
        if orderby_ascending is None:
            ValueError("Please specify if `orderby` is ascending or not.")
        if orderby_ascending:
            how = 'min'
        else:
            how = 'max'
        top_lrs = _aggregate_scores(liana_res, what=orderby, how=how,
                                    entities=['interaction',
                                              'ligand_complex',
                                              'receptor_complex']
                                    )
        top_lrs = top_lrs.sort_values('score', ascending=orderby_ascending).head(top_n).interaction
        # Filter liana_res to the interactions in top_lrs
        liana_res = liana_res[liana_res.interaction.isin(top_lrs)]
        
    # inverse sc if needed
    if inverse_colour:
        liana_res[colour] = _inverse_scores(liana_res[colour])
    if inverse_size:
        liana_res[size] = _inverse_scores(liana_res[size])

    # generate plot
    p = (ggplot(liana_res, aes(x='target', y='interaction', colour=colour, size=size))
         + geom_point()
         + facet_grid('~source')
         + scale_size_continuous(range=size_range)
         + labs(color=str.capitalize(colour),
                size=str.capitalize(size),
                y="Interactions (Ligand -> Receptor)",
                x="Target",
                title="Source")
         + theme_bw()
         + theme(legend_text=element_text(size=14),
                 strip_background=element_rect(fill="white"),
                 strip_text=element_text(size=15, colour="black"),
                 axis_text_y=element_text(size=10, colour="black"),
                 axis_title_y=element_text(colour="#808080", face="bold", size=15),
                 axis_text_x=element_text(size=11, face="bold", angle=90),
                 figure_size=figure_size,
                 plot_title=element_text(vjust=0, hjust=0.5, face="bold",
                                         colour="#808080", size=15)
                 )
         )

    if return_fig:
        return p

    p.draw()

    
def dotplot_by_sample(adata: anndata.AnnData  = None,
                      uns_key: str = 'liana_res',
                      liana_res: pandas.DataFrame =None,
                      sample_key: str = 'sample',
                      colour: str  = None,
                      size: str = None,
                      inverse_colour: bool = False,
                      inverse_size: bool = False,
                      source_labels: str | None =None,
                      target_labels: str | None =None,
                      ligand_complex: str | None =None, 
                      receptor_complex: str | None =None,
                      size_range: tuple = (2, 9),
                      figure_size: tuple = (8, 6),
                      return_fig: bool = True
                      ):
    """
    A dotplot of interactions by sample
    
    Parameters
    ----------
        adata
            adata object with liana_res and  in adata.uns. Defaults to None.
        uns_key
            key in adata.uns where liana_res is stored. Defaults to 'liana_res'.
        liana_res
            liana_res a DataFrame in liana's format. Defaults to None.
        sample_key
            sample_key used to group different samples/contexts from `liana_res`. Defaults to 'sample'.
        colour
            `column` in `liana_res` to define the colours of the dots. Defaults to None.
        size
            `column` in `liana_res` to define the size of the dots. Defaults to None.
        inverse_colour
            Whether to -log10 the `colour` column for plotting. `False` by default. Defaults to False.
        inverse_size
            Whether to -log10 the `size` column for plotting. `False` by default. Defaults to False.
        source_labels
            `list` with keys as `source` and values as `label` to be used in the plot. Defaults to None.
        target_labels
            `list` with keys as `target` and values as `label` to be used in the plot. Defaults to None.
        ligand_complex
            `list` of ligand complexes to filter the interactions to be plotted. Defaults to None.
        receptor_complex
            `list` of receptor complexes to filter the interactions to be plotted. Defaults to None.
        size_range
            Define size range - (min, max). Default is (2, 9). Defaults to (2, 9).
        figure_size
            Figure x,y size. Defaults to (8, 6).
        return_fig
            `bool` whether to return the fig object, `False` only plots. Defaults to True.
        
    Returns
    -------
    Returns a dotplot of Class ggplot for the specified interactions by sample.
    
    """
    
    liana_res = _prep_liana_res(adata=adata,
                                liana_res=liana_res, 
                                source_labels=source_labels,
                                target_labels=target_labels,
                                size=size,
                                colour=colour,
                                uns_key=uns_key)
    
    if ligand_complex is not None:
        liana_res = liana_res[np.isin(liana_res['ligand_complex'], ligand_complex)]
    if receptor_complex is not None:
        liana_res = liana_res[np.isin(liana_res['receptor_complex'], receptor_complex)]
        
        
    # inverse sc if needed
    if inverse_colour:
        liana_res[colour] = _inverse_scores(liana_res[colour])
    if inverse_size:
        liana_res[size] = _inverse_scores(liana_res[size])

    p = (ggplot(liana_res, aes(x='target', y='source', colour=colour, size=size))
            + geom_point()
            + facet_grid(f'interaction~{sample_key}', space='free', scales='free')
            + scale_size_continuous(range=size_range)
            + labs(color=str.capitalize(colour),
                   size=str.capitalize(size),
                   y="Source",
                   x="Target",
                   title=sample_key.capitalize())
            + theme_bw()
            + theme(legend_text=element_text(size=14),
                    strip_background=element_rect(fill="white"),
                    strip_text=element_text(size=13, colour="black", angle=90),
                    axis_text_y=element_text(size=10, colour="black"),
                    axis_title_y=element_text(colour="#808080", face="bold", size=12),
                    axis_text_x=element_text(size=11, face="bold", angle=90),
                    axis_title_x=element_text(colour="#808080", face="bold", size=12),
                    figure_size=figure_size,
                    plot_title=element_text(vjust=0, hjust=0.5, face="bold", size=12),
                    )
            )
    if return_fig:
        return p
    
    p.draw()


def _prep_liana_res(adata=None,
                    liana_res=None,
                    source_labels=None,
                    target_labels=None,
                    colour=None,
                    size=None,
                    uns_key='liana_res'):
    if colour is None:
        raise ValueError('`colour` must be provided!')
    if size is None:
        raise ValueError('`size` must be provided!')
    
    if (liana_res is None) & (adata is None):
        raise AttributeError(f'Ambiguous! One of `liana_res` or `adata.uns[{uns_key}]` should be provided.')
    if adata is not None:
        assert uns_key in adata.uns_keys()
        liana_res = adata.uns[uns_key].copy()
    if liana_res is not None:
        liana_res = liana_res.copy()
    if (liana_res is None) & (adata is None):
        raise ValueError('`liana_res` or `adata` must be provided!')

    # subset to only cell labels of interest
    liana_res = _filter_labels(liana_res, labels=source_labels, label_type='source')
    liana_res = _filter_labels(liana_res, labels=target_labels, label_type='target')
    
    liana_res['interaction'] = liana_res['ligand_complex'] + ' -> ' + liana_res['receptor_complex']

    return liana_res


def _aggregate_scores(res, what, how, entities):
    return res.groupby(entities).agg(score=(what, how)).reset_index()


def _inverse_scores(score):
    return -np.log10(score + np.finfo(float).eps)


def _filter_labels(liana_res, labels, label_type):
    if labels is not None:
        if labels is str:
            labels = [labels]
        covered = np.isin(labels, liana_res[label_type])
        if not covered.all():
            not_covered = np.array(labels)[~covered]
            raise ValueError(f"{not_covered} not found in `liana_res['{label_type}']`!")
        msk = np.isin(liana_res[label_type], labels)
        liana_res = liana_res[msk]
        
    return liana_res
        


from liana.method._Method import Method, MethodMeta
from .._pipe_utils._get_mean_perms import _get_lr_pvals


# simplified/resource-generalizable cellchat probability score
def _lr_probability(ligand_trimean, receptor_trimean):
    lr_prob = ligand_trimean * receptor_trimean
    return lr_prob / (0.5 + lr_prob)  # Kh=0.5


# Internal Function to calculate CellPhoneDB LR_mean and p-values
def _cellchat_score(x, perms, ligand_pos, receptor_pos, labels_pos) -> tuple:
    """
    Calculate CellChat-like LR means and p-values

    Parameters
    ----------
    x
        DataFrame row
    perms
        3D tensor with permuted averages per cluster
    ligand_pos
        Index of the ligand in the tensor
    receptor_pos
        Index of the receptor in the perms tensor
    labels_pos
        Index of cell identities in the perms tensor

    Returns
    -------
    A tuple with lr_mean and pvalue for x

    """
    return _get_lr_pvals(x=x,
                         perms=perms,
                         ligand_pos=ligand_pos,
                         receptor_pos=receptor_pos,
                         labels_pos=labels_pos,
                         agg_fun=_lr_probability,
                         ligand_col='ligand_trimean',
                         receptor_col='receptor_trimean')


# Initialize CellChat Meta
_cellchat = MethodMeta(method_name="CellChat",
                       complex_cols=["ligand_trimean", "receptor_trimean"],
                       add_cols=['mat_max'],
                       fun=_cellchat_score,
                       magnitude="lr_probs",
                       magnitude_ascending=False,
                       specificity="cellchat_pvals",
                       specificity_ascending=True,
                       permute=True,
                       reference="Jin, S., Guerrero-Juarez, C.F., Zhang, L., Chang, I., Ramos, "
                                 "R., Kuan, C.H., Myung, P., Plikus, M.V. and Nie, Q., "
                                 "2021. Inference and analysis of cell-cell communication using "
                                 "CellChat. Nature communications, 12(1), pp.1-20. "
                       )

# Initialize callable Method instance
cellchat = Method(_SCORE=_cellchat)



import scanpy as sc

def kang_2018():
    """
    Utility function to load the data from Kang et al., 2018; GSE96583. 
    The data contains ~25k PBMCs cells from 8 pooled patient lupus samples, each before and after IFN-beta stimulation.
    Kang, H., Subramaniam, M., Targ, S. et al. Multiplexed droplet single-cell RNA-sequencing using natural genetic variation. Nat Biotechnol 36, 89–94 (2018). https://doi.org/10.1038/nbt.4042
    
    Returns
    -------
    
    Returns a largely pre-processed AnnData object with the following attributes:
    Raw counts for ~25k cells; ~15k genes; 16 samples; 2 conditions.
    """
    ## TODO adapt when the CCC chapter one changes
    # load data as from CCC chapter
    adata = sc.read("kang_counts_25k.h5ad", backup_url="https://figshare.com/ndownloader/files/34464122")
    
    # Store the counts for later use
    adata.layers["counts"] = adata.X.copy()
    # Rename label to condition, replicate to patient
    adata.obs = adata.obs.rename({"label": "condition", "replicate": "patient"}, axis=1)
    
    # assign sample
    adata.obs["sample"] = (adata.obs["condition"].astype("str") + "&" + adata.obs["patient"].str.slice(8, 13))
    
    # set cell_types abbreviations (recommended given MOFA appends names)
    abbreviations = {'CD4 T cells':'CD4T',
                     'B cells':'B',
                     'NK cells':'NK',
                     'CD8 T cells':'CD8T',
                     'FCGR3A+ Monocytes':'FGR3',
                     'CD14+ Monocytes':'CD14',
                     'Dendritic cells':'DCs',
                     'Megakaryocytes':'Mega'}
    adata.obs["cell_abbr"] = adata.obs["cell_type"].replace(abbreviations)
    
    return adata


import pandas
from numpy import max, min

from liana.method import cellphonedb, singlecellsignalr as sca, \
    natmi, connectome, logfc, geometric_mean, cellchat
    
from liana.testing._toy_adata import get_toy_adata

# load toy adata
adata = get_toy_adata()
expected_shape = adata.shape


def test_cellchat():
    cellchat(adata, groupby='bulk_labels', use_raw=True, n_perms=2)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lr_probs' in liana_res.columns
    assert 'cellchat_pvals' in liana_res.columns
    assert max(liana_res[(liana_res.ligand == "TIMP1")].lr_probs) == 0.20561589810421071


def test_cellphonedb():
    cellphonedb(adata, groupby='bulk_labels', use_raw=True, n_perms=2)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lr_means' in liana_res.columns
    assert 'cellphone_pvals' in liana_res.columns
    assert max(liana_res[(liana_res.ligand == "TIMP1")].lr_means) == 2.134743630886078


def test_geometric_mean():
    geometric_mean(adata, groupby='bulk_labels', use_raw=True, n_perms=2)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lr_gmeans' in liana_res.columns
    assert 'gmean_pvals' in liana_res.columns
    assert max(liana_res[(liana_res.ligand == "TIMP1")].lr_gmeans) == 2.126363309240465


def test_natmi():
    natmi(adata, groupby='bulk_labels', use_raw=True)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'expr_prod' in liana_res.columns
    assert 'spec_weight' in liana_res.columns
    assert max(liana_res[(liana_res.ligand == "TIMP1")].spec_weight) == 0.0604750001773605
    assert max(liana_res[(liana_res.ligand == "TIMP1")].expr_prod) == 4.521420922884062


def test_sca():
    sca(adata, groupby='bulk_labels', use_raw=True)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lrscore' in liana_res.columns
    assert max(liana_res[(liana_res.ligand == "TIMP1")].lrscore) == 0.781133536891427


def test_logfc():
    logfc(adata, groupby='bulk_labels', use_raw=True)

    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'lr_logfc' in liana_res.columns
    assert max(liana_res[(liana_res.ligand == "TIMP1")].lr_logfc) == 1.4352725744247437


def test_connectome():
    connectome(adata, groupby='bulk_labels', use_raw=True)
    assert adata.shape == expected_shape
    assert 'liana_res' in adata.uns.keys()

    liana_res = adata.uns['liana_res']
    assert isinstance(liana_res, pandas.DataFrame)

    assert 'expr_prod' in liana_res.columns
    assert 'scaled_weight' in liana_res.columns
    assert max(liana_res[(liana_res.ligand == "TIMP1")].scaled_weight) == 0.9669451713562012
    assert max(liana_res[(liana_res.ligand == "TIMP1")].expr_prod) == 4.521420922884062


def test_with_all_lrs():
    natmi(adata, groupby='bulk_labels', use_raw=True, return_all_lrs=True, key_added='all_res')
    lr_all = adata.uns['all_res']
    assert lr_all.shape == (4200, 15)
    assert all(lr_all[~lr_all.lrs_to_keep][natmi.magnitude] == min(lr_all[natmi.magnitude])) is True
    assert all(lr_all[~lr_all.lrs_to_keep][natmi.specificity] == min(lr_all[natmi.specificity])) is True


def test_methods_by_sample():
    logfc.by_sample(adata, groupby='bulk_labels', use_raw=True, return_all_lrs=True, sample_key='sample')
    lr_by_sample = adata.uns['liana_res']
    
    assert 'sample' in lr_by_sample.columns
    assert lr_by_sample.shape == (10836, 15)
    


from ..multi import to_tensor_c2c, adata_to_views, lrs_to_views, get_variable_loadings, get_factor_scores
from ..testing import sample_lrs, get_toy_adata

import numpy as np
import pandas as pd

import cell2cell as c2c

adata = get_toy_adata()


def test_to_tensor_c2c():
    """Test to_tensor_c2c."""
    liana_res = sample_lrs(by_sample=True)
    
    liana_dict = to_tensor_c2c(liana_res=liana_res,
                               sample_key='sample',
                               score_key='specificity_rank',
                               return_dict=True
                               )
    assert isinstance(liana_dict, dict)
    
    tensor = to_tensor_c2c(liana_res=liana_res,
                           sample_key='sample',
                           score_key='specificity_rank')
    assert isinstance(tensor, c2c.tensor.tensor.PreBuiltTensor)
    assert tensor.sparsity_fraction()==0.0
    

def test_lrs_to_views():
    """Test lrs_to_views."""
    liana_res = sample_lrs(by_sample=True)
    adata.uns['liana_results'] = liana_res
    
    mdata = lrs_to_views(adata=adata,
                         sample_key='sample',
                         score_key='specificity_rank',
                         uns_key = 'liana_results',
                         obs_keys = ['case'],
                         lr_prop=0.1,
                         lrs_per_sample=0,
                         lrs_per_view=5,
                         samples_per_view=0,
                         min_variance=-1, # don't filter
                         verbose=True
                         )
    
    assert mdata.shape == (4, 16)
    assert 'case' in mdata.obs.columns
    assert len(mdata.varm_keys())==3

    

def test_adata_to_views():
    """Test adata_to_views."""
    mdata = adata_to_views(adata,
                           groupby='bulk_labels',
                           sample_key='sample',
                           obs_keys=None,
                           min_prop=0.05,
                           min_smpls=1,
                           min_cells=5,
                           min_counts=10,
                           mode='sum',
                           verbose=True,
                           use_raw=True,
                           skip_checks=True # skip because it's log-normalized (it's OK because toydata)
                           )
    
    assert len(mdata.varm_keys())==8
    assert 'case' not in mdata.obs.columns
    assert mdata.shape == (4, 5403)
    
    
    
def test_get_funs():
    liana_res = sample_lrs(by_sample=True)
    adata.uns['liana_results'] = liana_res
    
    mdata = lrs_to_views(adata=adata,
                         sample_key='sample',
                         score_key='specificity_rank',
                         uns_key = 'liana_results',
                         lr_prop=0.1,
                         lrs_per_sample=0,
                         lrs_per_view=5,
                         samples_per_view=0,
                         min_variance=-1, # don't filter
                         verbose=True
                         )
    
    # generate random loadings
    mdata.varm['LFs'] = np.random.rand(mdata.shape[1], 5)
    
    loadings = get_variable_loadings(mdata, 0, view_separator=':', variable_separator='^', pair_separator='&')
    assert isinstance(loadings, pd.DataFrame)
    assert loadings.shape == (16, 5)
    
    # dont drop columns & and don't separate
    loadings = get_variable_loadings(mdata, 0, drop_columns=False)
    assert isinstance(loadings, pd.DataFrame)
    assert loadings.shape == (16, 2)
    
    
    # generate random factor scores
    mdata.obsm['X_mofa'] = np.random.rand(mdata.shape[0], 5)
    
    scores = get_factor_scores(mdata)
    assert isinstance(scores, pd.DataFrame)
    assert scores.shape == (4, 6)



from setuptools import setup, find_packages
import os


def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='liana',
      version=get_version("liana/__init__.py"),
      description='LIANA - a LIgand-receptor ANalysis frAmework',
      author='Daniel Dimitrov',
      author_email='daniel.dimitrov@uni-heidelberg.de',
      url='https://github.com/saezlab/liana-py',
      project_urls={
          "Bug Tracker": "https://github.com/saezlab/liana-py/issues",
      },
      packages=find_packages(),
      package_data={'liana': ['resource/omni_resource.csv']},
      long_description=long_description,
      long_description_content_type="text/markdown",
      install_requires=["tqdm",
                        "pandas",
                        "anndata",
                        "scanpy",
                        "plotnine"
                        ],
      python_requires=">=3.7",
      classifiers=[
          "Programming Language :: Python :: 3",
          "Operating System :: OS Independent"]
      )




import numpy as np
from liana.plotting import dotplot, dotplot_by_sample
from liana.testing import sample_lrs

liana_res = sample_lrs()


def test_check_dotplot_order():
    my_p = dotplot(liana_res=liana_res,
                   size='specificity_rank',
                   colour='magnitude',
                   top_n=20,
                   orderby='specificity_rank',
                   orderby_ascending=False,
                   target_labels=["A", "B", "C"]
                   )
    assert my_p is not None
    assert 'interaction' in my_p.data.columns
    np.testing.assert_equal(np.unique(my_p.data.interaction).shape, (20,))
    set(my_p.data.target)
    assert {'A', 'B', 'C'} == set(my_p.data.target)


def test_check_doplot_filter():
    my_p2 = dotplot(liana_res=liana_res,
                    size='specificity_rank',
                    colour='magnitude',
                    filterby='specificity_rank',
                    filter_lambda=lambda x: x > 0.95,
                    inverse_colour=True,
                    source_labels=["A"]
                    )
    assert my_p2 is not None
    # we force this, but not intended all interactions
    # to be only 0.95, but rather for an interaction to get
    # plotted, in at least one cell type pair it should be > 0.95
    assert all(my_p2.data['specificity_rank'] > 0.95) is True


def test_dotplot_bysample():
    liana_res = sample_lrs(by_sample=True)
    my_p3 = dotplot_by_sample(liana_res=liana_res,
                              size='specificity_rank',
                              colour='magnitude',
                              target_labels='E',
                              sample_key='sample')
    assert my_p3 is not None
    assert 'interaction' in my_p3.data.columns
    assert 'sample' in my_p3.data.columns
    assert 'B' not in my_p3.data['target']
    
    


from ._Method import Method, MethodMeta, _show_methods
from .sc._rank_aggregate import AggregateClass, _rank_aggregate_meta
from .sc import cellphonedb, connectome, logfc, natmi, singlecellsignalr, geometric_mean, cellchat

import numpy as np

# callable consensus instance
_methods = [cellphonedb, connectome, logfc, natmi, singlecellsignalr, cellchat]
rank_aggregate = AggregateClass(_rank_aggregate_meta, methods=_methods)


def show_methods():
    """Shows methods available in LIANA"""
    return _show_methods(_methods + [rank_aggregate, geometric_mean])

def get_method_scores():
    """Returns a dict of all scoring functions, with a boolean indicating whether the score is ascending or not"""
    instances = np.array(MethodMeta.instances)
    relevant = np.array([(isinstance(instance, Method)) | (isinstance(instance, AggregateClass)) for instance in instances])
    instances = instances[relevant]
    specificity_scores = {method.specificity: method.specificity_ascending for method in instances if method.specificity is not None}
    magnitude_scores = {method.magnitude : method.magnitude_ascending for method in instances if method.magnitude is not None}
    scores = {**specificity_scores, **magnitude_scores}
    return scores



from scipy.stats import gmean

from liana.method._Method import Method, MethodMeta
from .._pipe_utils._get_mean_perms import _get_lr_pvals


def _gmean(x, y):
    return gmean([x, y], axis=0)


# Internal Function to calculate Geometric LR_mean and p-values
def _gmean_score(x, perms, ligand_pos, receptor_pos, labels_pos) -> tuple:
    """
    Calculate CellPhoneDB-like LR means and p-values

    Parameters
    ----------
    x
        DataFrame row
    perms
        3D tensor with permuted averages per cluster
    ligand_pos
        Index of the ligand in the tensor
    receptor_pos
        Index of the receptor in the perms tensor
    labels_pos
        Index of cell identities in the perms tensor

    Returns
    -------
    A tuple with lr_mean and pvalue for x

    """
    return _get_lr_pvals(x, perms, ligand_pos, receptor_pos, labels_pos, _gmean)


# Initialize Meta
_geometric_mean = MethodMeta(method_name="Geometric Mean",
                             complex_cols=["ligand_means", "receptor_means"],
                             add_cols=[],
                             fun=_gmean_score,
                             magnitude="lr_gmeans",
                             magnitude_ascending=False,
                             specificity="gmean_pvals",
                             specificity_ascending=True,
                             permute=True,
                             reference="CellPhoneDBv2's permutation approach applied to the "
                                       "geometric means of ligand-receptors' mean, as opposed to "
                                       "their arithmetic mean."
                             )

# Initialize callable Method instance
geometric_mean = Method(_SCORE=_geometric_mean)
